import time

from openpectus.lang.exec.tags import Tag, TagDirection, TagFormatFunction
from openpectus.lang.exec.events import BlockInfo

# Make sure the mark separator does not conflict with ArchiverTag delimiter.
# This wold make the archiver unable to write its archive file.
# MARK_SEPARATOR = ", "
MARK_SEPARATOR = "; "


class ReadingTag(Tag):
    """ Represents a common reading, i.e. a input tag with float values. """
    def __init__(self, name: str, unit: str | None = None, format_fn: TagFormatFunction | None = None) -> None:
        super().__init__(name, value=0.0, unit=unit, direction=TagDirection.Input, format_fn=format_fn)

    def set_value(self, val, *args, **kwargs):
        if val is not None:
            try:
                float(val)
            except Exception:
                raise ValueError(
                    f'ReadingTag "{self.name}" cannot be set to "{val}" because it cannot be coerced to float type.')
        super().set_value(val, *args, **kwargs)


class SelectTag(Tag):
    """ Represents a tag with choice values. """
    def __init__(self, name: str, value, unit: str | None,
                 choices: list[str], direction: TagDirection = TagDirection.NA) -> None:
        super().__init__(name=name, value=value, unit=unit, direction=direction)
        if choices is None or len(choices) == 0:
            raise ValueError(f'SelectTag "{self.name}" has no choices. Choices must be non-empty.')
        self.choices = choices

    def set_value(self, val, *args, **kwargs):
        assert val in self.choices, \
            f'SelectTag "{self.name}" cannot be set to "{val}" because it is not among the valid options: "{self.choices}".'
        super().set_value(val, *args, **kwargs)


class MarkTag(Tag):
    def __init__(self) -> None:
        super().__init__(name="Mark")

    def set_value(self, val: int | float | str | None, tick_time: float, *args, **kwargs) -> None:
        """ Append value to existing value """
        assert isinstance(val, str), f"The Mark tag value must be a str, not a {type(val).__name__}"
        value = str(self.get_value() or "")
        if value == "":
            value = str(val)
        else:
            value += MARK_SEPARATOR + str(val)
        super().set_value(value, tick_time, *args, **kwargs)

    def archive(self) -> str | None:
        """ Reset value and return it """
        value = self.value
        super().set_value("", time.time())
        return str(value or "")


class AccumulatorTag(Tag):
    """ Generic accumulator tag. Can be used as the system tag
        SystemTagName.ACCUMULATED_VOLUME.
    """
    def __init__(self, name: str, totalizer: Tag):
        super().__init__(name)
        self.totalizer: Tag = totalizer
        self.unit = self.totalizer.unit
        self.v0: float | None = None

    def reset(self):
        self.v0 = self.totalizer.as_float()
        self.value = 0.0
        assert self.v0 is not None

    def on_start(self, run_id: str):
        self.reset()

    def on_tick(self, tick_time: float, increment_time: float):
        assert self.v0 is not None, f"Error in aggregator tag '{self.name}', v0 was not set."
        self.value = self.totalizer.as_float() - self.v0


class AccumulatorBlockTag(Tag):
    """ Implements a block accumulator. Can be used as the system tag
        SystemTagName.BLOCK_VOLUME.
    """
    def __init__(self, name: str, totalizer: Tag):
        super().__init__(name)
        self.totalizer: Tag = totalizer
        self.root_block = BlockInfo("", 0)
        self.accumulator_stack: list[AccumulatorTag] = []
        self.accumulator_stack.append(AccumulatorTag(self.root_block.key, self.totalizer))
        self.cur_block = self.root_block
        self.cur_accumulator = self.accumulator_stack[0]

        self.value = 0.0
        self.unit = self.totalizer.unit

    def on_start(self, run_id: str):
        self.cur_accumulator.on_start(run_id)

    def on_tick(self, tick_time: float, increment_time: float):
        # tick all accumulators in scope to update block values
        for acc in self.accumulator_stack:
            acc.on_tick(tick_time, increment_time)

        # apply the current value
        self.value = self.cur_accumulator.get_value()

    def on_block_start(self, block_info: BlockInfo):
        self.cur_block = block_info
        self.cur_accumulator = AccumulatorTag(self.cur_block.key, self.totalizer)
        self.accumulator_stack.append(self.cur_accumulator)

        assert self.run_id is not None, f"Error in AccumulatorBlockTag tag '{self.name}', run_id not set in on_block_start."
        self.cur_accumulator.on_start(self.run_id)

    def on_block_end(self, block_info: BlockInfo, new_block_info: BlockInfo | None):
        if len(self.accumulator_stack) == 0:
            raise ValueError("Stack is empty. This means on_block_end was called in error")
        self.accumulator_stack.pop()
        self.cur_accumulator = self.accumulator_stack[-1]


class AccumulatedColumnVolume(Tag):
    def __init__(self, name: str, column_volume: Tag, totalizer: Tag):
        super().__init__(name)
        self.totalizer: Tag = totalizer
        self.column_volume: Tag = column_volume
        self.unit = "CV"
        self.v0 = 0.0

    def reset(self):
        self.v0 = self.totalizer.as_float()
        self.value = 0.0
        assert self.v0 is not None

    def on_start(self, run_id: str):
        self.reset()

    def on_tick(self, tick_time: float, increment_time: float):
        cv = self.column_volume.as_float()
        v = self.totalizer.as_float()
        if cv == 0.0:
            self.value = None
        else:
            self.value = (v-self.v0) / cv



import logging
import unittest
import threading
import queue
import asyncio
import asyncua
import asyncua.ua
import asyncua.common.callback
from typing import Any, Dict
import os
from openpectus.engine.hardware import Register, RegisterDirection, HardwareLayerException
from openpectus.engine.opcua_hardware import OPCUA_Hardware, OPCUA_Types

logging.basicConfig(format=" %(name)s :: %(levelname)-8s :: %(message)s")
logger = logging.getLogger("openpectus.engine.opcua_hardware.OPCUA_Hardware")
logger.setLevel(logging.DEBUG)

opcua_host = "opc.tcp://127.0.0.1:484{:02d}"


async def opcua_test_server_task(port_suffix: int,
                                 registers: Dict[str, Register],
                                 callback_output_queue,
                                 stop_event,
                                 is_started_event,
                                 is_finished_event) -> None:
    post_write_callback_queue = queue.Queue()
    server = asyncua.Server()
    await server.init()
    server.set_endpoint(opcua_host.format(port_suffix))
    server.set_server_name("Open Pectus OPC-UA Test Server")
    server.set_security_policy([asyncua.ua.SecurityPolicyType.NoSecurity,])
    uri = "https://github.com/Open-Pectus/Open-Pectus/"

    idx = await server.register_namespace(uri)
    objects = server.nodes.objects

    for r in registers.values():
        # Objects/2:parent_node/2:node_name -> ["Objects", "2:parent_node", "2:node_name"]
        breadcrumbs = r.options["path"].split("/")
        # ["Objects", "2:parent_node", "2:node_name"] -> ["2:parent_node",]
        objects_prior_to_variable = breadcrumbs[1:-1]
        obj = objects
        if objects_prior_to_variable:
            for part in objects_prior_to_variable:
                idx, name = part.split(":")
                obj = await obj.add_object(int(idx), name)
        idx, name = breadcrumbs[-1].split(":")
        node = await obj.add_variable(int(idx),
                                      name,
                                      0.0 if "type" not in r.options else
                                      asyncua.ua.Variant([], r.options["type"])
                                      )
        if RegisterDirection.Write in r.direction:
            await node.set_writable()

    await server.start()

    def post_write_callback(event, dispatcher):
        for write_value in event.request_params.NodesToWrite:
            post_write_callback_queue.put(write_value)
    server.subscribe_server_callback(asyncua.common.callback.CallbackType.PostWrite, post_write_callback)

    while not stop_event.is_set():
        await asyncio.sleep(0.1)
        if not is_started_event.is_set():
            is_started_event.set()
    server.unsubscribe_server_callback(asyncua.common.callback.CallbackType.PostWrite, post_write_callback)

    while not post_write_callback_queue.empty():
        write_value = post_write_callback_queue.get()
        node = server.get_node(write_value.NodeId)
        node_path_list = await node.get_path(as_string=True)
        # ["0:Root", "0:Objects", 2:"node_name"] -> "Objects/2:node_name"
        node_path = "/".join(node_path_list[1:])[2:]
        tag_value = write_value.Value.Value.Value
        callback_output_queue.put((node_path, tag_value,))
    await server.stop()
    await asyncio.gather(*asyncio.all_tasks() - {asyncio.current_task()})
    is_finished_event.set()


def sync_opcua_test_server_task(*args, **kwargs):
    asyncio.run(opcua_test_server_task(*args, **kwargs))


class OPCUATestServer():
    """ Context manager which launches OPC-UA test server. """
    def __init__(self, port_suffix: int = 0, registers=None):
        self.port_suffix = port_suffix
        self.callback_output_queue = queue.Queue()
        self.server_write_events: list[tuple[str, Any]] = []
        self.registers = dict()
        if registers:
            self.registers = registers

    def __enter__(self):
        self.stop_event = threading.Event()  # Signal to OPC-UA test server that it should stop
        self.is_started_event = threading.Event()  # Signal from OPC-UA test server that it has started
        self.is_finished_event = threading.Event()  # Signal from OPC-UA test server that it has finished
        self.process = threading.Thread(
            target=sync_opcua_test_server_task,
            args=(
                self.port_suffix,
                self.registers,
                self.callback_output_queue,
                self.stop_event,
                self.is_started_event,
                self.is_finished_event,
            ),
            daemon=True
        )
        self.process.start()
        # Block until OPC-UA server is started
        self.is_started_event.wait()

    def __exit__(self, type, value, traceback):
        self.stop_event.set()
        self.is_finished_event.wait()
        # Gather queue before closing the process to avoid corruption
        while not self.callback_output_queue.empty():
            self.server_write_events.append(self.callback_output_queue.get())
        self.process.join(timeout=2)


@unittest.skipIf(bool(os.environ.get("OPENPECTUS_ENGINE_SKIP_SLOW_TESTS")), reason="Skipping slow tests as configured")
class TestOPCUAHardware(unittest.TestCase):
    def test_can_connect(self):
        port_suffix = 0
        test_server = OPCUATestServer(port_suffix)
        with test_server:
            hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
            try:
                hwl.connect()
                hwl.disconnect()
            except HardwareLayerException:
                self.fail()

    def test_reconnect(self):
        port_suffix = 1
        FT01 = Register("FT01", RegisterDirection.Read, path="Objects/2:FT01")
        registers = {"FT01": FT01}

        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = registers

        # Connect and read
        with OPCUATestServer(port_suffix, registers=registers):
            hwl.connect()
            self.assertEqual(0.0, hwl.read(FT01))
        # Sever connection to server and read again
        with self.assertRaises(HardwareLayerException):
            hwl.read(FT01)
        # Re-onnect and read
        with OPCUATestServer(port_suffix, registers=registers), hwl:
            self.assertEqual(0.0, hwl.read(FT01))

    def test_connection_left_connected_does_not_hang(self):
        port_suffix = 2
        FT01 = Register("FT01", RegisterDirection.Read, path="Objects/2:FT01")
        registers = {"FT01": FT01}
        test_server = OPCUATestServer(port_suffix, registers=registers)

        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = registers
        with test_server:
            hwl.connect()  # Connection is opened but not closed subsequently

    def test_can_read_register(self):
        port_suffix = 3
        FT01 = Register("FT01", RegisterDirection.Read, path="Objects/2:FT01")
        registers = {"FT01": FT01}
        test_server = OPCUATestServer(port_suffix, registers=registers)

        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = registers
        with test_server, hwl:
            self.assertEqual(0.0, hwl.read(FT01))

    def test_can_write_register(self):
        port_suffix = 4
        PU01 = Register("PU01", RegisterDirection.Both, path="Objects/2:PU01")
        registers = {"PU01": PU01}
        test_server = OPCUATestServer(port_suffix, registers=registers)

        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = registers
        with test_server, hwl:
            hwl.write(10.2, PU01)
        self.assertIn((PU01.options["path"], 10.2,), test_server.server_write_events)

    def test_can_read_multiple_registers(self):
        port_suffix = 5
        registers = {f"Readable {i:02d}": Register(f"{i:02d}",
                                                   RegisterDirection.Read,
                                                   path=f"Objects/2:readable_{i:02d}") for i in range(100)}
        test_server = OPCUATestServer(port_suffix, registers=registers)

        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = registers
        with test_server, hwl:
            values = hwl.read_batch(list(registers.values()))
            for value in values:
                self.assertEqual(0.0, value)

    def test_can_write_multiple_registers(self):
        port_suffix = 6
        registers = {f"Writable {i:02d}": Register(f"{i:02d}",
                     RegisterDirection.Both,
                     path=f"Objects/2:writable_{i:02d}") for i in range(100)}
        test_server = OPCUATestServer(port_suffix, registers=registers)
        values_to_write = [float(i) for i in range(len(registers))]

        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = registers
        with test_server, hwl:
            hwl.write_batch(values_to_write, list(registers.values()))
        for register, value_to_write in zip(registers.values(), values_to_write):
            self.assertIn((register.options["path"], value_to_write,), test_server.server_write_events)

    def test_does_help_user_fix_path_typo(self):
        port_suffix = 7
        FT01_totalizer_1_total_with_typo = Register("FT01",
                                                    RegisterDirection.Read,
                                                    path="Objects/2:FT01/2:Totalizer/2:Totaal 1")
        FT01_totalizer_1_total_proper = Register("FT01",
                                                 RegisterDirection.Read,
                                                 path="Objects/2:FT01/2:Totalizer/2:Total 1")
        test_server = OPCUATestServer(port_suffix, registers={"FT01": FT01_totalizer_1_total_proper})

        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = {"FT01": FT01_totalizer_1_total_with_typo}
        with test_server, hwl:
            with self.assertRaises(HardwareLayerException) as context:
                hwl.read(FT01_totalizer_1_total_with_typo)
            self.assertIn("Objects/2:FT01/2:Totalizer/2:Totaal 1 cannot be found", str(context.exception))

    def test_can_read_write_registers_with_type(self):
        port_suffix = 8
        bool_register = Register("Bool",
                                 RegisterDirection.Both,
                                 path="Objects/2:Boolean value",
                                 type=OPCUA_Types.Boolean)
        float_register = Register("Float",
                                  RegisterDirection.Both,
                                  path="Objects/2:Float value",
                                  type=OPCUA_Types.Float)
        uint16_register = Register("Uint16",
                                   RegisterDirection.Both,
                                   path="Objects/2:Uint16 value",
                                   type=OPCUA_Types.UInt16)
        registers = {
            "Bool": bool_register,
            "Float": float_register,
            "Uint16": uint16_register,
        }
        test_server = OPCUATestServer(port_suffix, registers=registers)

        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = registers

        with test_server, hwl:
            hwl.write_batch([1, 1, 1], list(registers.values()))
            values = hwl.read_batch(list(registers.values()))

        self.assertIn((bool_register.options["path"], True), test_server.server_write_events)
        self.assertIn((float_register.options["path"], 1.0), test_server.server_write_events)
        self.assertIn((uint16_register.options["path"], 1), test_server.server_write_events)
        self.assertEqual(values, [True, 1.0, 1,])

    def test_illegal_register_type_is_detected(self):
        port_suffix = 9
        FT01_type_not_ok = Register("FT01", RegisterDirection.Read, path="Objects/2:FT01", type=bool)
        FT01_type_ok = Register("FT01", RegisterDirection.Read, path="Objects/2:FT01", type=OPCUA_Types.Boolean)
        test_server = OPCUATestServer(port_suffix, registers={"FT01": FT01_type_ok})

        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = {"FT01": FT01_type_not_ok}
        with test_server, hwl:
            with self.assertRaises(HardwareLayerException):
                hwl.validate_offline()

    def test_type_disparity_is_detected(self):
        port_suffix = 10
        bool_register = Register("Bool",
                                 RegisterDirection.Both,
                                 path="Objects/2:Boolean value",
                                 type=OPCUA_Types.Boolean)
        registers = {
            "Bool": bool_register,
        }
        test_server = OPCUATestServer(port_suffix, registers=registers)

        bool_register = Register("Bool",
                                 RegisterDirection.Both,
                                 path="Objects/2:Boolean value",
                                 type=OPCUA_Types.Float)
        registers = {
            "Bool": bool_register,
        }
        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = registers
        with test_server, hwl:
            with self.assertRaises(HardwareLayerException):
                hwl.validate_online()

    def test_access_level_disparity_is_detected(self):
        port_suffix = 11
        FT01 = Register("FT01", RegisterDirection.Read, path="Objects/2:FT01")
        registers = {"FT01": FT01}
        test_server = OPCUATestServer(port_suffix, registers=registers)

        FT01 = Register("FT01", RegisterDirection.Write, path="Objects/2:FT01")
        registers = {"FT01": FT01}
        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = registers
        with test_server, hwl:
            with self.assertRaises(HardwareLayerException):
                hwl.validate_online()

    def test_exception_when_attempting_to_access_using_disconnected_client(self):
        port_suffix = 12
        FT01 = Register("FT01", RegisterDirection.Read, path="Objects/2:FT01")
        registers = {"FT01": FT01}
        test_server = OPCUATestServer(port_suffix, registers=registers)

        hwl = OPCUA_Hardware(host=opcua_host.format(port_suffix))
        hwl._registers = registers
        with test_server:
            hwl.connect()
            hwl.read(FT01)
            hwl.disconnect()
            with self.assertRaises(HardwareLayerException) as context:
                hwl.read(FT01)
            self.assertIn("OPC-UA client is closed.", str(context.exception))


def attempt_to_use_hardware():
    registers = {r.name: r for r in
                 [
                    Register("Readable",
                             RegisterDirection.Read,
                             path="Objects/3:ServerInterfaces/4:OPCUA/4:Readable",
                             type=OPCUA_Types.Float),
                    Register("Writable",
                             RegisterDirection.Write,
                             path="Objects/3:ServerInterfaces/4:OPCUA/4:Writable",
                             type=OPCUA_Types.Float),
                    Register("ReadableAndWritable",
                             RegisterDirection.Both,
                             path="Objects/3:ServerInterfaces/4:OPCUA/4:ReadableAndWritable",
                             type=OPCUA_Types.Float),
                    Register("Bool",
                             RegisterDirection.Both,
                             path="Objects/3:ServerInterfaces/4:OPCUA/4:UserDataType/4:Bool",
                             type=OPCUA_Types.Boolean),
                    Register("Float",
                             RegisterDirection.Both,
                             path="Objects/3:ServerInterfaces/4:OPCUA/4:UserDataType/4:Float",
                             type=OPCUA_Types.Float),
                    Register("Int",
                             RegisterDirection.Both,
                             path="Objects/3:ServerInterfaces/4:OPCUA/4:UserDataType/4:Int",
                             type=OPCUA_Types.Int16),
                    Register("Uint",
                             RegisterDirection.Both,
                             path="Objects/3:ServerInterfaces/4:OPCUA/4:UserDataType/4:Uint",
                             type=OPCUA_Types.UInt16),
                 ]
                 }

    hwl = OPCUA_Hardware(host="opc.tcp://192.168.0.1:4840/")
    hwl._registers = registers
    hwl.validate_offline()
    with hwl:
        hwl.validate_online()
        for r in [r for r in hwl.registers.values() if RegisterDirection.Read in r.direction]:
            print(r, hwl.read(r))
        hwl.write(-5.0, registers["Writable"])
        hwl.write(-6.0, registers["ReadableAndWritable"])
        hwl.write(5, registers["Bool"])
        hwl.write(1.02, registers["Float"])
        hwl.write(-54000, registers["Int"])
        hwl.write(-54, registers["Uint"])


if __name__ == "__main__":
    unittest.main()



import unittest

from openpectus.lang.grammar.pgrammar import PGrammar
from openpectus.lang.grammar.pprogramformatter import PProgramFormatter, print_program
from openpectus.lang.model.pprogram import (
    PCommandWithDuration,
    PDuration,
    PNode,
    PProgram,
    PBlank,
    PBlock,
    PEndBlock,
    PEndBlocks,
    PWatch,
    PAlarm,
    PMark,
    PCommand,
    PComment,
    PErrorInstruction,
    PCondition
)


def build(s):
    p = PGrammar()
    p.parse(s)
    return p


def node_missing_start_position_count(node: PNode) -> int:
    count = 0
    if node.line is None:
        count += 1
    if node.indent is None:
        count += 1
    for n in node.get_child_nodes(recursive=True):
        count += node_missing_start_position_count(n)
    return count


class BuilderTest(unittest.TestCase):
    def assertProgramMatches(self, program: PProgram, expected: str):
        expected = expected.strip()
        out = PProgramFormatter().format(program)
        self.assertMultiLineEqual(expected, out)

    def assertNodeHasType(self, node: PNode, expected_type: type):
        if not isinstance(node, expected_type):
            raise AssertionError(f"Invalid node type: {type(node).__name__}, expected: {expected_type.__name__}")

    def test_command(self):
        p = build("foo: bar=baz")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)

        command: PCommand = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(command)
        self.assertEqual("foo", command.name)
        self.assertEqual("bar=baz", command.args)
        # print_program(program)

    def test_start(self):
        p = build("Start")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        command: PCommand = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(command)
        self.assertEqual("Start", command.name)

    def test_increment_run_counter(self):
        p = build("Increment run counter")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        command: PCommand = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(command)
        self.assertEqual("Increment run counter", command.name)

    def test_run_counter(self):
        p = build("Run counter: 4")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        command: PCommand = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(command)
        self.assertEqual("Run counter", command.name)
        self.assertEqual("4", command.args)

    def test_block(self):
        p = build("Block: foo")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        block: PBlock = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(block)
        self.assertEqual("foo", block.name)
        # print_program(program)

    def test_block_with_end_block(self):
        p = build(
            """Block: foo
    End block"""
        )
        program = p.build_model()
        # p.printSyntaxTree(p.tree)

        block: PBlock = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(block)
        self.assertEqual("foo", block.name)
        # print_program(program)

        end_block = block.get_child_nodes()[0]
        self.assertIsInstance(end_block, PEndBlock)

    def test_block_with_invalid_indentation(self):
        p = build(" Block: foo")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)

        block: PBlock = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(block)
        # print_program(program)
        self.assertIsInstance(block, PBlock)
        self.assertTrue(block.has_error())

    def test_block_with_invalid_scope_indentation(self):
        p = build("    Block: foo")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)

        block: PBlock = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(block)
        # print_program(program)
        self.assertIsInstance(block, PBlock)
        self.assertTrue(block.has_error())

    def test_block_with_end_block_with_invalid_indentation(self):
        p = build(
            """Block: foo
 End block"""
        )
        program = p.build_model()
        # p.printSyntaxTree(p.tree)

        block: PBlock = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(block)
        self.assertEqual("foo", block.name)
        # print_program(program)

        end_block = block.get_child_nodes()[0]
        self.assertIsInstance(end_block, PEndBlock)
        self.assertTrue(end_block.has_error())

    def test_block_with_end_block_with_missing_indentation(self):
        p = build(
            """Block: foo
End block"""
        )
        program = p.build_model()

        block: PBlock = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(block)
        self.assertEqual("foo", block.name)
        # print_program(program, show_errors=True)

        # TODO clarify - do we allow empty blocks?
        # Note: This may not be an error. It is an empty Block, followed
        # by an End block which happens to do nothing. An analyzer
        # might issue a warning about it.
        # The current implementation disallows empty blocks and thus
        # gives an indentation error on End block

        end_block = block.get_child_nodes()[0]
        self.assertIsInstance(end_block, PEndBlock)
        self.assertTrue(end_block.has_error())

    def test_block_with_mark_with_missing_indentation(self):
        p = build(
            """Block: foo
Mark: foo"""
        )
        program = p.build_model()
        # p.printSyntaxTree(p.tree)

        block: PBlock = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(block)
        self.assertEqual("foo", block.name)
        # print_program(program, show_errors=True)

        mark = block.get_child_nodes()[0]
        self.assertIsInstance(mark, PMark)
        self.assertTrue(mark.has_error())

        # TODO: Note that Mark in contained in Block. May or may not be what we want in this case.

    def test_block_with_command_and_end_block(self):
        p = build(
            """Block: foo
    cmd: args
    End block"""
        )
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program)

        block: PBlock = program.get_instructions()[0]  # type: ignore
        self.assertIsNotNone(block)
        self.assertEqual("foo", block.name)

        children = block.get_child_nodes()
        assert isinstance(children[0], PCommand)
        cmd = children[0]
        self.assertEqual("cmd", cmd.name)
        self.assertEqual("args", cmd.args)

        assert isinstance(children[1], PEndBlock)

    def test_block_with_command(self):
        p = build(
            """Block: Equilibration
    0.96 Inlet PU01: VA01"""
        )
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program)

        instructions = program.get_instructions()
        # TODO should program have both instructions? probably not. may have an iterator though
        self.assertEqual(2, len(instructions))
        self.assertIsInstance(instructions[0], PBlock)
        self.assertIsInstance(instructions[1], PCommand)

        block: PBlock = instructions[0]  # type: ignore
        self.assertEqual("Equilibration", block.name)
        self.assertEqual(0, block.indent)
        self.assertIsNone(block.time)
        self.assertFalse(block.has_error())

        cmd1: PCommand = instructions[1]  # type: ignore
        self.assertEqual("Inlet PU01", cmd1.name)
        self.assertEqual("VA01", cmd1.args)
        self.assertEqual(4, cmd1.indent)
        self.assertEqual(0.96, cmd1.time)
        self.assertFalse(cmd1.has_error())

    def test_block_Equilibration_partial(self):
        p = build(
            """Block: Equilibration
    0.0 Inlet PU01: VA01
    0.0 Inlet PU02: VA06
    0.0 Inlet PU04: VA15
    0.0 PU01: 77.64 %
    0.0 PU02: 10.00 %
    0.0 PU04: 12.36 %"""
        )
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program)

        instructions = program.get_instructions()
        self.assertEqual(7, len(instructions))
        self.assertIsInstance(instructions[0], PBlock)
        for i in [1, 2, 3, 4, 5]:
            self.assertIsInstance(instructions[i], PCommand)

        block: PBlock = instructions[0]  # type: ignore
        self.assertEqual("Equilibration", block.name)
        self.assertEqual(0, block.indent)
        self.assertIsNone(block.time)

        cmd1: PCommand = instructions[1]  # type: ignore
        self.assertEqual("Inlet PU01", cmd1.name)
        self.assertEqual("VA01", cmd1.args)
        self.assertEqual(4, cmd1.indent)
        self.assertEqual(0.0, cmd1.time)
        self.assertFalse(cmd1.has_error())

        cmd2: PCommand = instructions[2]  # type: ignore
        self.assertEqual("Inlet PU02", cmd2.name)
        self.assertEqual("VA06", cmd2.args)
        self.assertEqual(4, cmd2.indent)
        self.assertEqual(0.0, cmd2.time)
        self.assertFalse(cmd2.has_error())

    def test_block_Equilibration(self):
        p = build("""
Block: Equilibration
    0.0 Inlet PU01: VA01
    0.0 Inlet PU02: VA06
    0.0 Inlet PU04: VA15
    0.0 PU01: 77.64 %
    0.0 PU02: 10.00 %
    0.0 PU04: 12.36 %
    2.95 Zero UV
    3.0 End block
    """)

        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program)

        # check wiki for updated expressions
        instructions = program.get_instructions()
        self.assertIsInstance(instructions[0], PBlock)
        for i in [1, 2, 3, 4, 5, 6, 7]:
            instr = instructions[i]
            self.assertIsInstance(instr, PCommand)
            self.assertFalse(instr.has_error())
        instr = instructions[8]
        self.assertIsInstance(instr, PEndBlock)
        self.assertFalse(instr.has_error())

    def test_block_name_non_identifier(self):
        p = build("Block: 87")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program)
        block = program.get_instructions()[0]
        assert isinstance(block, PBlock)
        self.assertEqual(block.name, "87")
        self.assertFalse(program.has_error(recursive=True))

    def test_end_blocks(self):
        p = build("End blocks")
        program = p.build_model()
        instructions = program.get_instructions()
        self.assertIsInstance(instructions[0], PEndBlocks)

    def test_mark(self):
        p = build("Mark: A")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program)
        mark = program.get_instructions()[0]
        assert isinstance(mark, PMark)
        self.assertEqual(mark.name, "A")
        self.assertFalse(program.has_error(recursive=True))

    def test_mark_comment(self):
        p = build("Mark: A # foo")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program)
        mark = program.get_instructions()[0]
        self.assertIsInstance(mark, PMark)
        self.assertEqual(mark.comment, " foo")
        self.assertFalse(program.has_error(recursive=True))

    def test_mark_name_non_identifier(self):
        def test(code: str, name: str):
            with self.subTest(code):
                p = build(code)
                program = p.build_model()
                # p.printSyntaxTree(p.tree)
                # print_program(program)
                mark = program.get_instructions()[0]
                assert isinstance(mark, PMark)
                self.assertEqual(mark.name, name)
                self.assertFalse(program.has_error(recursive=True))

        test("Mark: 87", "87")
        test("Mark: 5 seconds have passed", "5 seconds have passed")
        test("Mark: B %", "B %")

    def test_indentation_error(self):
        code = """
    Mark: A
Mark: B"""
        # test that an indentation error (A) does not interfere with the scope
        # of following instructions (B)
        p = build(code)
        program = p.build_model()
        [a, b] = program.get_instructions()
        # print_program(program, show_errors=True)
        self.assertTrue(a.has_error())
        self.assertFalse(b.has_error())

    def test_blanks(self):
        p = build(
            """
Mark: A
        """
        )
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program)

        non_blanks = program.get_instructions()
        self.assertEqual(1, len(non_blanks))
        self.assertIsInstance(non_blanks[0], PMark)

        all = program.get_instructions(include_blanks=True)
        self.assertEqual(3, len(all))
        self.assertFalse(program.has_error(recursive=True))

    def test_sequential_blocks(self):
        p = build(
            """
Block: A1
    Mark: A1 Start
    End block
Block: A2
    Mark: A2 Start
    End block
        """
        )
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program)
        self.assertFalse(program.has_error(recursive=True))
        self.assertEqual(0, node_missing_start_position_count(program))

        self.assertProgramMatches(
            program,
            """
PProgram
    PBlock
        PMark
        PEndBlock
    PBlock
        PMark
        PEndBlock
        """,
        )

    def test_nested_blocks(self):
        p = build(
            """
Block: A
    Mark: A Start
    Block: A1
        Mark: A1 Start
        End block
    Block: A2
        Mark: A2 Start
        End block
    Mark: A End
    End block
        """
        )
        program = p.build_model()

        # print_program(program, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))
        self.assertEqual(0, node_missing_start_position_count(program))

        self.assertProgramMatches(
            program,
            """
PProgram
    PBlock
        PMark
        PBlock
            PMark
            PEndBlock
        PBlock
            PMark
            PEndBlock
        PMark
        PEndBlock
        """,
        )

    def test_outdent(self):
        p = build(
            """
Block: A1
    Mark: A1 Start
Mark: foo
"""
        )
        program = p.build_model()

        # print_program(program)
        self.assertFalse(program.has_error(recursive=True))
        self.assertEqual(0, node_missing_start_position_count(program))

        self.assertProgramMatches(
            program,
            """
PProgram
    PBlock
        PMark
    PMark
""",
        )

    def test_outdent_multiple(self):
        p = build(
            """
Block: A
    Mark: A Start
    Block: A1
        Mark: A1 Start
Mark: foo
"""
        )
        program = p.build_model()

        # print_program(program)
        self.assertFalse(program.has_error(recursive=True))
        self.assertEqual(0, node_missing_start_position_count(program))

        self.assertProgramMatches(
            program,
            """
PProgram
    PBlock
        PMark
        PBlock
            PMark
    PMark
""",
        )

    def test_commands(self):
        p = build(
            """
Incr_counter
Incr counter
Incr_counter: foo
Incr counter: foo
Incr_counter: foo=bar
Incr counter: foo='bar',2
"""
        )
        # Note: grammer seems to require cmd_args with colon - otherwise the following line break causes error
        # This may be a problem when typing live
        program = p.build_model()
        # print_program(program)
        cmds = program.get_commands()
        self.assertEqual(6, len(cmds))

    def test_watch(self):
        p = build(
            """
Watch: Run counter > 0
    Mark: A"""
        )
        program = p.build_model()
        [watch, mark_a] = program.get_instructions(include_blanks=False)
        self.assertNodeHasType(watch, PWatch)
        self.assertNodeHasType(mark_a, PMark)
        self.assertFalse(watch.has_error(recursive=True))
        self.assertIsNone(watch._inst_error)

    def test_watch_missing_condition(self):
        p = build(
            """
Watch
    Mark: A"""
        )
        program = p.build_model()
        [watch, mark] = program.get_instructions(include_blanks=False)

        self.assertNodeHasType(watch, PWatch)
        self.assertIsNone(watch._inst_error)
        self.assertNodeHasType(mark, PMark)
        self.assertTrue(watch.has_error())

    def test_watch_partial_condition(self):
        p = build(
            """
Watch:
    Mark: A"""
        )
        program = p.build_model()
        [watch, mark] = program.get_instructions(include_blanks=False)
        self.assertNodeHasType(watch, PWatch)
        self.assertNodeHasType(mark, PMark)
        self.assertIsNotNone(watch._inst_error)
        self.assertTrue(watch.has_error())

    def test_watch_partial_condition_w_comment(self):
        p = build(
            """
Watch: # foo
    Mark: A"""
        )
        program = p.build_model()
        [watch, mark] = program.get_instructions(include_blanks=False)
        self.assertNodeHasType(watch, PWatch)
        self.assertEqual(watch.comment, " foo")
        self.assertNodeHasType(mark, PMark)
        self.assertIsNotNone(watch._inst_error)
        self.assertTrue(watch.has_error())

    def test_watch_unit(self):
        p = build(
            """
Mark: a
Watch: counter > 0 mL
    Mark: b
Mark: c
incr counter
Watch: counter > 0 mL
    Mark: d
        """
        )
        program = p.build_model()

        # p.printSyntaxTree(p.tree)
        # print_program(program, show_line_numbers=True, show_errors=True)
        self.assertFalse(program.has_error(recursive=True))
        non_blanks = program.get_instructions()
        self.assertIsInstance(non_blanks[0], PMark)
        self.assertIsInstance(non_blanks[1], PWatch)
        self.assertIsInstance(non_blanks[1].get_child_nodes()[0], PMark)
        assert isinstance(non_blanks[1], PWatch)
        condition = non_blanks[1].condition
        assert condition is not None
        self.assertEqual("counter", condition.tag_name)
        self.assertEqual(">", condition.op)
        self.assertEqual("0", condition.tag_value)
        self.assertEqual("mL", condition.tag_unit)

    def test_watch_unit_percent(self):
        p = build(
            """
Watch: counter > 0 %
    Mark: A
"""
        )
        program = p.build_model()

        # p.printSyntaxTree(p.tree)
        # print_program(program, show_line_numbers=True, show_errors=True)
        self.assertFalse(program.has_error(recursive=True))
        non_blanks = program.get_instructions()
        self.assertIsInstance(non_blanks[0], PWatch)
        self.assertIsInstance(non_blanks[0].get_child_nodes()[0], PMark)
        assert isinstance(non_blanks[0], PWatch)
        condition = non_blanks[0].condition
        assert condition is not None
        self.assertEqual("counter", condition.tag_name)
        self.assertEqual(">", condition.op)
        self.assertEqual("0", condition.tag_value)
        self.assertEqual("%", condition.tag_unit)

    def test_watch_unitless(self):
        p = build(
            """
Mark: a
Watch: counter > 0
    Mark: b
Mark: c
incr counter
Watch: counter > 0
    Mark: d
        """
        )
        program = p.build_model()

        # p.printSyntaxTree(p.tree)
        # print_program(program, show_line_numbers=True, show_errors=True)
        self.assertFalse(program.has_error(recursive=True))
        non_blanks = program.get_instructions()
        self.assertIsInstance(non_blanks[0], PMark)
        self.assertIsInstance(non_blanks[1], PWatch)
        self.assertIsInstance(non_blanks[1].get_child_nodes()[0], PMark)
        assert isinstance(non_blanks[1], PWatch)
        condition = non_blanks[1].condition
        assert condition is not None
        self.assertEqual("counter", condition.tag_name)
        self.assertEqual(">", condition.op)
        self.assertEqual("0", condition.tag_value)
        self.assertEqual(None, condition.tag_unit)

    def test_navigation(self):
        p = build(
            """
Mark: a
Watch: counter > 0
    Mark: b
Mark: c
incr counter
Watch: counter > 0 ml
    Mark: d

        """
        )

        program = p.build_model()

        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))
        all = program.get_instructions(include_blanks=True)

        blank_1, mark_a, watch_1, mark_b, mark_c, incr, watch_2, mark_d, blank_2, blank_3 = all

        self.assertIsInstance(blank_1, PBlank)
        self.assertEqual(1, blank_1.line)

        self.assertIsInstance(mark_a, PMark)
        self.assertEqual(2, mark_a.line)

        self.assertIsInstance(watch_1, PWatch)
        self.assertEqual(3, watch_1.line)

        self.assertIsInstance(mark_b, PMark)
        self.assertEqual(4, mark_b.line)

        self.assertIsInstance(mark_c, PMark)
        self.assertEqual(5, mark_c.line)

        self.assertIsInstance(incr, PCommand)
        self.assertEqual(6, incr.line)

        self.assertIsInstance(watch_2, PWatch)
        self.assertEqual(7, watch_2.line)

        self.assertIsInstance(mark_d, PMark)
        self.assertEqual(8, mark_d.line)

        self.assertIsInstance(blank_2, PBlank)
        self.assertEqual(9, blank_2.line)

        self.assertIsInstance(blank_3, PBlank)
        self.assertEqual(10, blank_3.line)

    def test_alarm(self):
        p = build(
            """
Mark: A
Alarm: counter > 0
    Mark: b
Mark: c
Increment counter
Mark: d
""")
        program = p.build_model()

        self.assertFalse(program.has_error(recursive=True))
        n_mark_a, n_alarm, n_mark_b, n_mark_c, n_increment, n_mark_d = program.get_instructions()
        self.assertIsInstance(n_alarm, PAlarm)
        self.assertIsInstance(n_mark_a, PMark)
        self.assertIsInstance(n_mark_b, PMark)
        self.assertIsInstance(n_mark_c, PMark)
        self.assertIsInstance(n_mark_d, PMark)
        self.assertIsInstance(n_increment, PCommand)
        self.assertTrue(n_alarm.get_child_nodes()[0], n_mark_b)

        assert isinstance(n_alarm, PAlarm)
        condition = n_alarm.condition
        assert condition is not None
        self.assertEqual("counter", condition.tag_name)
        self.assertEqual(">", condition.op)
        self.assertEqual("0", condition.tag_value)
        self.assertEqual(None, condition.tag_unit)

    def test_program_mark(self):
        p = build(
            """
Mark:  a
        """
        )

        program = p.build_model()
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))

    def test_program_watch(self):
        p = build(
            """
Watch A > 0     # missing colon
    Mark: a
        """
        )

        program = p.build_model()
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertTrue(program.has_error(recursive=True))

    def test_command_comment(self):
        p = build(
            """
mycommand: a  # foo  
        """
        )

        program = p.build_model()
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))

        all = program.get_instructions(include_blanks=True)
        blank_1, cmd, blank_2 = all
        self.assertIsInstance(cmd, PCommand)
        self.assertEqual(" foo  ", cmd.comment)

    def test_comment(self):
        p = build(
            """
# foo
# bar # baz
        """
        )

        program = p.build_model()
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))

        all = program.get_instructions(include_blanks=True)

        blank_1, comment, comment2, blank_2 = all

        self.assertIsInstance(blank_1, PBlank)
        self.assertEqual(1, blank_1.line)

        self.assertIsInstance(comment, PComment)
        self.assertEqual(2, comment.line)
        self.assertEqual(" foo", comment.comment)

        self.assertIsInstance(comment2, PComment)
        self.assertEqual(3, comment2.line)
        # slight mishap here. we ignore it for now
        self.assertEqual(" bar# baz", comment2.comment)
        # self.assertEqual(" bar # baz", comment2.comment)

    def test_restart(self):
        p = build(
            """
Mark: A
Restart
        """
        )

        program = p.build_model()
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))

        all = program.get_instructions(include_blanks=True)
        blank_1, mark, restart, blank_2 = all

        self.assertIsInstance(blank_1, PBlank)
        self.assertEqual(1, blank_1.line)

        self.assertIsInstance(mark, PMark)
        self.assertIsInstance(restart, PCommand)
        assert isinstance(restart, PCommand)
        self.assertEqual(restart.name, "Restart")

    def test_stop(self):
        p = build(
            """
Mark: A
Stop
        """
        )

        program = p.build_model()
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))

        all = program.get_instructions(include_blanks=True)
        blank_1, mark, stop, blank_2 = all

        self.assertIsInstance(blank_1, PBlank)
        self.assertEqual(1, blank_1.line)

        self.assertIsInstance(mark, PMark)
        self.assertIsInstance(stop, PCommand)
        assert isinstance(stop,  PCommand)
        self.assertEqual(stop.name, "Stop")

    def test_pause(self):
        p = build("Pause")
        program = p.build_model()
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))

        [pause] = program.get_instructions(include_blanks=True)

        self.assertIsInstance(pause, PCommandWithDuration)
        assert isinstance(pause,  PCommandWithDuration)
        self.assertEqual(pause.name, "Pause")
        self.assertEqual(pause.duration, None)

    def test_pause_w_arg(self):
        p = build("Pause: 2 min")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))

        [pause] = program.get_instructions(include_blanks=True)

        self.assertIsInstance(pause, PCommandWithDuration)
        assert isinstance(pause,  PCommandWithDuration)
        self.assertEqual(pause.name, "Pause")
        assert isinstance(pause.duration,  PDuration)
        self.assertEqual(pause.duration.time, 2.0)
        self.assertEqual(pause.duration.unit, "min")


    def test_hold(self):
        p = build("Hold")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))

        [hold] = program.get_instructions(include_blanks=True)

        self.assertIsInstance(hold, PCommandWithDuration)
        assert isinstance(hold,  PCommandWithDuration)
        self.assertEqual(hold.name, "Hold")
        self.assertEqual(hold.duration, None)

    def test_hold_w_arg(self):
        p = build("Hold: 27.35 s")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))

        [hold] = program.get_instructions(include_blanks=True)

        self.assertIsInstance(hold, PCommandWithDuration)
        assert isinstance(hold,  PCommandWithDuration)
        self.assertEqual(hold.name, "Hold")
        assert isinstance(hold.duration,  PDuration)
        self.assertEqual(hold.duration.time, 27.35)
        self.assertEqual(hold.duration.unit, "s")

    def test_hold_w_arg_error(self):
        p = build("Hold: 27.35")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertTrue(program.has_error(recursive=True))

        [hold] = program.get_instructions(include_blanks=True)

        self.assertIsInstance(hold, PCommandWithDuration)
        assert isinstance(hold,  PCommandWithDuration)
        self.assertEqual(hold.name, "Hold")
        assert isinstance(hold.duration,  PDuration)
        self.assertTrue(hold.duration.error)
        self.assertEqual(hold.duration.time, None)
        self.assertEqual(hold.duration.unit, None)

    def test_wait_must_have_duration(self):
        p = build("Wait")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertTrue(program.has_error(recursive=True))

    def test_wait_w_arg(self):
        p = build("Wait:2h")
        program = p.build_model()
        # p.printSyntaxTree(p.tree)
        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))

        [wait] = program.get_instructions(include_blanks=True)

        self.assertIsInstance(wait, PCommandWithDuration)
        assert isinstance(wait,  PCommandWithDuration)
        self.assertEqual(wait.name, "Wait")
        assert isinstance(wait.duration,  PDuration)
        self.assertEqual(wait.duration.time, 2.0)
        self.assertEqual(wait.duration.unit, "h")

    def test_info_warning_error(self):
        p = build("""
Info:foo
Warning: bar
Error: baz
""")
        program = p.build_model()
        self.assertFalse(program.has_error(recursive=True))
        [info, warning, error] = program.get_instructions()
        assert isinstance(info, PCommand)
        self.assertEqual(info.args, "foo")
        assert isinstance(warning, PCommand)
        self.assertEqual(warning.args, "bar")
        assert isinstance(error, PCommand)
        self.assertEqual(error.args, "baz")


    @unittest.skip("Need a better error concept for this to make sense")
    def test_program_errors(self):
        p = build(
            """
Mark:  a  
Mark  b  # comment
Watch: counter >  
    Mark: c
Bad command
:
        """
        )

        program = p.build_model()

        # print_program(program, show_blanks=True, show_errors=True, show_line_numbers=True)
        self.assertFalse(program.has_error(recursive=True))
        all = program.get_instructions(include_blanks=True)

        blank_1, mark_a, mark_b, watch, mark_c, bad, err, blank_2 = all

        self.assertIsInstance(blank_1, PBlank)
        self.assertEqual(1, blank_1.line)

        self.assertIsInstance(mark_a, PMark)
        self.assertEqual(2, mark_a.line)

        # Note: this malformed Mark "Mark a" is parsed as command
        # It will still fail command name validation later on, though
        self.assertIsInstance(mark_b, PCommand)
        # self.assertIsInstance(mark_b, PErrorInstruction)
        self.assertEqual(3, mark_b.line)

        self.assertIsInstance(watch, PWatch)
        self.assertEqual(4, watch.line)
        assert isinstance(watch, PWatch)
        assert isinstance(watch.condition, PCondition)
        self.assertTrue(watch.condition.error)

        self.assertIsInstance(mark_c, PMark)
        self.assertEqual(5, mark_c.line)

        self.assertIsInstance(bad, PCommand)
        self.assertEqual(6, bad.line)

        self.assertIsInstance(err, PErrorInstruction)
        self.assertEqual(7, err.line)


if __name__ == "__main__":
    unittest.main()



import re
import unittest

from openpectus.lang.exec.uod import RegexNamedArgumentParser
from openpectus.engine.uod_builder_api import (
    RegexCategorical, RegexNumber, RegexText,
    as_float, as_int
)


class TestRegexs_named_groups(unittest.TestCase):

    def test_named_groups_RegexNumberWithUnit(self):
        parser = RegexNamedArgumentParser(RegexNumber(units=None))
        self.assertEqual(['number'], parser.get_named_groups())
        parser = RegexNamedArgumentParser(RegexNumber(units=["kg"]))
        self.assertEqual(["number", "number_unit"], parser.get_named_groups())

    def test_get_units_RegexNumberWithUnit(self):
        # single unit
        parser = RegexNamedArgumentParser(RegexNumber(units=['kg']))
        self.assertEqual(['kg'], parser.get_units())
        # multiple units
        parser = RegexNamedArgumentParser(RegexNumber(units=["kg", "g", "m2"]))
        self.assertEqual(["kg", "g", "m2"], parser.get_units())

    def test_named_groups_RegexText(self):
        parser = RegexNamedArgumentParser(RegexText())
        self.assertEqual(["text"], parser.get_named_groups())

    def test_named_groups_RegexCategorical(self):
        # exclusive
        regex = RegexCategorical(exclusive_options=["Open", "Closed"])
        parser = RegexNamedArgumentParser(regex)
        self.assertEqual(['option'], parser.get_named_groups())

        # exclusive + additive
        regex = RegexCategorical(exclusive_options=['Closed'], additive_options=['VA01', 'VA02', 'VA03'])
        parser = RegexNamedArgumentParser(regex)
        self.assertEqual(['option'], parser.get_named_groups())


class TestRegexs(unittest.TestCase):
    def test_regex_number_with_unit(self):
        regex = RegexNumber(units=["mS/cm", "L/h", "bar"])

        self.assertEqual(
            re.search(regex, "12 L/h").groupdict(),  # type: ignore
            dict(number="12", number_unit="L/h"))

        self.assertEqual(
            re.search(regex, "12.1 L/h").groupdict(),  # type: ignore
            dict(number="12.1", number_unit="L/h"))

        self.assertEqual(
            re.search(regex, ".5 L/h").groupdict(),  # type: ignore
            dict(number=".5", number_unit="L/h"))

        self.assertEqual(
            re.search(regex, "0.921 bar").groupdict(),  # type: ignore
            dict(number="0.921", number_unit="bar"))

        self.assertEqual(
            re.search(regex, "-1 bar").groupdict(),  # type: ignore
            dict(number="-1", number_unit="bar"))

        self.assertEqual(
            re.search(regex, "-12.1 bar").groupdict(),  # type: ignore
            dict(number="-12.1", number_unit="bar"))

        self.assertEqual(
            re.search(regex, "-.1 bar").groupdict(),  # type: ignore
            dict(number="-.1", number_unit="bar"))

        self.assertEqual(
            re.search(regex, "0.14 foo"),
            None)

    def test_regex_number_ws(self):
        regex = RegexNumber(units=["mS/cm", "L/h", "bar"])
        self.assertEqual(
            re.search(regex, " 12  L/h   ").groupdict(),  # type: ignore
            dict(number="12", number_unit="L/h"))

        regex = RegexNumber(units=None)
        self.assertEqual(
            re.search(regex, " 12  ").groupdict(),  # type: ignore
            dict(number="12"))


    def test_regex_number_with_unit_non_negative(self):
        regex = RegexNumber(units=['mS/cm', 'L/h', 'bar'], non_negative=True)

        self.assertEqual(
            re.search(regex, "12 L/h").groupdict(),  # type: ignore
            dict(number="12", number_unit="L/h"))

        self.assertEqual(
            re.search(regex, "12.1 L/h").groupdict(),  # type: ignore
            dict(number="12.1", number_unit="L/h"))

        self.assertEqual(
            re.search(regex, ".5 L/h").groupdict(),  # type: ignore
            dict(number=".5", number_unit="L/h"))

        self.assertEqual(
            re.search(regex, "0.921 bar").groupdict(),  # type: ignore
            dict(number="0.921", number_unit="bar"))

        self.assertEqual(
            re.search(regex, "-12 L/h"),
            None)

        self.assertEqual(
            re.search(regex, "0.14 foo"),
            None)

    def test_regex_number_without_unit(self):
        regex = RegexNumber(units=None)

        self.assertEqual(
            re.search(regex, "12").groupdict(),  # type: ignore
            dict(number="12"))

        self.assertEqual(
            re.search(regex, "12.1").groupdict(),  # type: ignore
            dict(number="12.1"))

        self.assertEqual(
            re.search(regex, ".5").groupdict(),  # type: ignore
            dict(number=".5"))

        self.assertEqual(
            re.search(regex, "-1").groupdict(),  # type: ignore
            dict(number="-1"))

        self.assertEqual(
            re.search(regex, "-12.1").groupdict(),  # type: ignore
            dict(number="-12.1"))

        self.assertEqual(
            re.search(regex, "-.1").groupdict(),  # type: ignore
            dict(number="-.1"))

        self.assertEqual(
            re.search(regex, "0.14 foo"),
            None)


    def test_regex_categorical_exclusive(self):
        regex = RegexCategorical(exclusive_options=["Open", "Closed"])
        self.assertEqual(
            re.search(regex, "Closed").groupdict(),  # type: ignore
            dict(option="Closed"))

        self.assertEqual(
            re.search(regex, "VA01"),
            None)

        self.assertEqual(
            re.search(regex, "Open").groupdict(),  # type: ignore
            dict(option="Open"))

        self.assertEqual(
            re.search(regex, "Open+Closed"),
            None)


    def test_regex_categorical_exclusive_and_additive(self):
        regex = RegexCategorical(exclusive_options=['Closed'], additive_options=['VA01', 'VA02', 'VA03'])

        self.assertEqual(
            re.search(regex, "Closed").groupdict(),  # type: ignore
            dict(option="Closed"))

        self.assertEqual(
            re.search(regex, "Closed+VA01"),
            None)

        self.assertEqual(
            re.search(regex, "VA01").groupdict(),  # type: ignore
            dict(option="VA01"))

        self.assertEqual(
            re.search(regex, "VA01+VA02").groupdict(),  # type: ignore
            dict(option="VA01+VA02"))

        self.assertEqual(
            re.search(regex, "Open+Closed"),
            None)


    def test_regex_string(self):
        regex = RegexText()

        self.assertEqual(
            re.search(regex, "Test string").groupdict(),  # type: ignore
            dict(text="Test string"))

        self.assertEqual(
            re.search(regex, "Test").groupdict(),  # type: ignore
            dict(text="Test"))

        self.assertEqual(
            re.search(regex, ""),
            None)

    def test_regex_string_allowing_empty(self):
        regex = RegexText(allow_empty=True)

        self.assertEqual(
            re.search(regex, "Test string").groupdict(),  # type: ignore
            dict(text="Test string"))

        self.assertEqual(
            re.search(regex, "Test").groupdict(),  # type: ignore
            dict(text="Test"))

        self.assertEqual(
            re.search(regex, "").groupdict(),  # type: ignore
            dict(text=""))


class TestConversions(unittest.TestCase):

    def test_as_float(self):
        def test_value(value: str, expected_result: float | None):
            with self.subTest(value):
                result = as_float(value)
                self.assertEqual(result, expected_result)

        test_value("0",  0)
        test_value("0.002",  0.002)
        test_value("-34",  -34)
        test_value("-37.31",  -37.31)
        test_value("0.0",  0.0)
        test_value("  3  ",  3.0)

        test_value("",  None)
        test_value("3f",  None)
        test_value("f",  None)

    def test_as_int(self):
        def test_value(value: str, expected_result: int | None):
            with self.subTest(value):
                result = as_int(value)
                self.assertEqual(result, expected_result)

        test_value("0",  0)
        test_value("3",  3)
        test_value("-34",  -34)
        test_value("   4  ",  4)

        test_value("-37.31",  None)
        test_value("",  None)
        test_value("3.2",  None)


if __name__ == "__main__":
    unittest.main()




import threading
import time
from typing import Callable


TickConsumer = Callable[[float, float], None]
""" Tick function. It is called by the timer on each tick with arguments:
- tick_time: float          Time of the tick in seconds
- increment_time: float     Time since last tick_time

Will be called on all ticks, even tick 0 where `increment_time` is zero.

On pause/resume, will also be called on the first tick after resume where `increment_time` is also zero.
"""

class EngineTimer():

    def set_tick_fn(self, tick_fn: TickConsumer):
        pass

    def start(self):
        pass

    def stop(self):
        pass


class NullTimer(EngineTimer):
    pass


class OneThreadTimer(EngineTimer):
    """ Single threaded (1 extra thread) timer.

    This allows controlled multithreading.
    """
    def __init__(self, interval: float, tick: TickConsumer | None = None) -> None:
        """ Initialize a new timer.

        Parameters:
        - interval: float   Interval in seconds between ticks
        - tick: function    The tick function to call when interval has passed
        """
        self.interval = interval
        self.tick = tick
        self.running = False            # flag to allow shut down by exiting the ticker loop method

    def set_tick_fn(self, tick_fn: TickConsumer):
        self.tick = tick_fn

    def start(self):
        """ Start timer """
        if self.tick is None:
            raise ValueError("Timer cannot start when no tick_fn has been set")
        self.running = True
        self.thread = threading.Thread(
            target=self.ticker,
            name=self.__class__.__name__,
            daemon=True)
        self.thread.start()

    def ticker(self):
        assert self.tick is not None, "No tick function has been set"

        tick_number = 0
        last_tick_time_mon = 0.0
        increment = 0.0

        while self.running:
            tick_time = time.time()
            tick_time_mon = time.monotonic()
            if tick_number > 0:
                increment = tick_time_mon - last_tick_time_mon

            self.tick(tick_time, increment)

            tick_number += 1
            last_tick_time_mon = tick_time_mon
            tick_time_mon = time.monotonic()
            elapsed = tick_time_mon - last_tick_time_mon
            deadline = self.interval - elapsed
            if deadline > 0.0:
                time.sleep(deadline)

    def stop(self):
        self.running = False



# Generated from c:/Projects/Novo/Open-Pectus/openpectus/lang/grammar/pcode.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .pcodeParser import pcodeParser
else:
    from pcodeParser import pcodeParser

# This class defines a complete listener for a parse tree produced by pcodeParser.
class pcodeListener(ParseTreeListener):

    # Enter a parse tree produced by pcodeParser#program.
    def enterProgram(self, ctx:pcodeParser.ProgramContext):
        pass

    # Exit a parse tree produced by pcodeParser#program.
    def exitProgram(self, ctx:pcodeParser.ProgramContext):
        pass


    # Enter a parse tree produced by pcodeParser#instruction_line.
    def enterInstruction_line(self, ctx:pcodeParser.Instruction_lineContext):
        pass

    # Exit a parse tree produced by pcodeParser#instruction_line.
    def exitInstruction_line(self, ctx:pcodeParser.Instruction_lineContext):
        pass


    # Enter a parse tree produced by pcodeParser#instruction.
    def enterInstruction(self, ctx:pcodeParser.InstructionContext):
        pass

    # Exit a parse tree produced by pcodeParser#instruction.
    def exitInstruction(self, ctx:pcodeParser.InstructionContext):
        pass


    # Enter a parse tree produced by pcodeParser#block.
    def enterBlock(self, ctx:pcodeParser.BlockContext):
        pass

    # Exit a parse tree produced by pcodeParser#block.
    def exitBlock(self, ctx:pcodeParser.BlockContext):
        pass


    # Enter a parse tree produced by pcodeParser#block_name.
    def enterBlock_name(self, ctx:pcodeParser.Block_nameContext):
        pass

    # Exit a parse tree produced by pcodeParser#block_name.
    def exitBlock_name(self, ctx:pcodeParser.Block_nameContext):
        pass


    # Enter a parse tree produced by pcodeParser#end_block.
    def enterEnd_block(self, ctx:pcodeParser.End_blockContext):
        pass

    # Exit a parse tree produced by pcodeParser#end_block.
    def exitEnd_block(self, ctx:pcodeParser.End_blockContext):
        pass


    # Enter a parse tree produced by pcodeParser#end_blocks.
    def enterEnd_blocks(self, ctx:pcodeParser.End_blocksContext):
        pass

    # Exit a parse tree produced by pcodeParser#end_blocks.
    def exitEnd_blocks(self, ctx:pcodeParser.End_blocksContext):
        pass


    # Enter a parse tree produced by pcodeParser#watch.
    def enterWatch(self, ctx:pcodeParser.WatchContext):
        pass

    # Exit a parse tree produced by pcodeParser#watch.
    def exitWatch(self, ctx:pcodeParser.WatchContext):
        pass


    # Enter a parse tree produced by pcodeParser#alarm.
    def enterAlarm(self, ctx:pcodeParser.AlarmContext):
        pass

    # Exit a parse tree produced by pcodeParser#alarm.
    def exitAlarm(self, ctx:pcodeParser.AlarmContext):
        pass


    # Enter a parse tree produced by pcodeParser#condition.
    def enterCondition(self, ctx:pcodeParser.ConditionContext):
        pass

    # Exit a parse tree produced by pcodeParser#condition.
    def exitCondition(self, ctx:pcodeParser.ConditionContext):
        pass


    # Enter a parse tree produced by pcodeParser#compare_op.
    def enterCompare_op(self, ctx:pcodeParser.Compare_opContext):
        pass

    # Exit a parse tree produced by pcodeParser#compare_op.
    def exitCompare_op(self, ctx:pcodeParser.Compare_opContext):
        pass


    # Enter a parse tree produced by pcodeParser#condition_lhs.
    def enterCondition_lhs(self, ctx:pcodeParser.Condition_lhsContext):
        pass

    # Exit a parse tree produced by pcodeParser#condition_lhs.
    def exitCondition_lhs(self, ctx:pcodeParser.Condition_lhsContext):
        pass


    # Enter a parse tree produced by pcodeParser#condition_rhs.
    def enterCondition_rhs(self, ctx:pcodeParser.Condition_rhsContext):
        pass

    # Exit a parse tree produced by pcodeParser#condition_rhs.
    def exitCondition_rhs(self, ctx:pcodeParser.Condition_rhsContext):
        pass


    # Enter a parse tree produced by pcodeParser#increment_rc.
    def enterIncrement_rc(self, ctx:pcodeParser.Increment_rcContext):
        pass

    # Exit a parse tree produced by pcodeParser#increment_rc.
    def exitIncrement_rc(self, ctx:pcodeParser.Increment_rcContext):
        pass


    # Enter a parse tree produced by pcodeParser#restart.
    def enterRestart(self, ctx:pcodeParser.RestartContext):
        pass

    # Exit a parse tree produced by pcodeParser#restart.
    def exitRestart(self, ctx:pcodeParser.RestartContext):
        pass


    # Enter a parse tree produced by pcodeParser#stop.
    def enterStop(self, ctx:pcodeParser.StopContext):
        pass

    # Exit a parse tree produced by pcodeParser#stop.
    def exitStop(self, ctx:pcodeParser.StopContext):
        pass


    # Enter a parse tree produced by pcodeParser#pause.
    def enterPause(self, ctx:pcodeParser.PauseContext):
        pass

    # Exit a parse tree produced by pcodeParser#pause.
    def exitPause(self, ctx:pcodeParser.PauseContext):
        pass


    # Enter a parse tree produced by pcodeParser#hold.
    def enterHold(self, ctx:pcodeParser.HoldContext):
        pass

    # Exit a parse tree produced by pcodeParser#hold.
    def exitHold(self, ctx:pcodeParser.HoldContext):
        pass


    # Enter a parse tree produced by pcodeParser#wait.
    def enterWait(self, ctx:pcodeParser.WaitContext):
        pass

    # Exit a parse tree produced by pcodeParser#wait.
    def exitWait(self, ctx:pcodeParser.WaitContext):
        pass


    # Enter a parse tree produced by pcodeParser#duration.
    def enterDuration(self, ctx:pcodeParser.DurationContext):
        pass

    # Exit a parse tree produced by pcodeParser#duration.
    def exitDuration(self, ctx:pcodeParser.DurationContext):
        pass


    # Enter a parse tree produced by pcodeParser#mark.
    def enterMark(self, ctx:pcodeParser.MarkContext):
        pass

    # Exit a parse tree produced by pcodeParser#mark.
    def exitMark(self, ctx:pcodeParser.MarkContext):
        pass


    # Enter a parse tree produced by pcodeParser#mark_name.
    def enterMark_name(self, ctx:pcodeParser.Mark_nameContext):
        pass

    # Exit a parse tree produced by pcodeParser#mark_name.
    def exitMark_name(self, ctx:pcodeParser.Mark_nameContext):
        pass


    # Enter a parse tree produced by pcodeParser#time.
    def enterTime(self, ctx:pcodeParser.TimeContext):
        pass

    # Exit a parse tree produced by pcodeParser#time.
    def exitTime(self, ctx:pcodeParser.TimeContext):
        pass


    # Enter a parse tree produced by pcodeParser#timeexp.
    def enterTimeexp(self, ctx:pcodeParser.TimeexpContext):
        pass

    # Exit a parse tree produced by pcodeParser#timeexp.
    def exitTimeexp(self, ctx:pcodeParser.TimeexpContext):
        pass


    # Enter a parse tree produced by pcodeParser#comment.
    def enterComment(self, ctx:pcodeParser.CommentContext):
        pass

    # Exit a parse tree produced by pcodeParser#comment.
    def exitComment(self, ctx:pcodeParser.CommentContext):
        pass


    # Enter a parse tree produced by pcodeParser#comment_text.
    def enterComment_text(self, ctx:pcodeParser.Comment_textContext):
        pass

    # Exit a parse tree produced by pcodeParser#comment_text.
    def exitComment_text(self, ctx:pcodeParser.Comment_textContext):
        pass


    # Enter a parse tree produced by pcodeParser#blank.
    def enterBlank(self, ctx:pcodeParser.BlankContext):
        pass

    # Exit a parse tree produced by pcodeParser#blank.
    def exitBlank(self, ctx:pcodeParser.BlankContext):
        pass


    # Enter a parse tree produced by pcodeParser#command.
    def enterCommand(self, ctx:pcodeParser.CommandContext):
        pass

    # Exit a parse tree produced by pcodeParser#command.
    def exitCommand(self, ctx:pcodeParser.CommandContext):
        pass


    # Enter a parse tree produced by pcodeParser#command_name.
    def enterCommand_name(self, ctx:pcodeParser.Command_nameContext):
        pass

    # Exit a parse tree produced by pcodeParser#command_name.
    def exitCommand_name(self, ctx:pcodeParser.Command_nameContext):
        pass


    # Enter a parse tree produced by pcodeParser#command_args.
    def enterCommand_args(self, ctx:pcodeParser.Command_argsContext):
        pass

    # Exit a parse tree produced by pcodeParser#command_args.
    def exitCommand_args(self, ctx:pcodeParser.Command_argsContext):
        pass


    # Enter a parse tree produced by pcodeParser#identifier.
    def enterIdentifier(self, ctx:pcodeParser.IdentifierContext):
        pass

    # Exit a parse tree produced by pcodeParser#identifier.
    def exitIdentifier(self, ctx:pcodeParser.IdentifierContext):
        pass


    # Enter a parse tree produced by pcodeParser#identifier_ext.
    def enterIdentifier_ext(self, ctx:pcodeParser.Identifier_extContext):
        pass

    # Exit a parse tree produced by pcodeParser#identifier_ext.
    def exitIdentifier_ext(self, ctx:pcodeParser.Identifier_extContext):
        pass


    # Enter a parse tree produced by pcodeParser#inst_error.
    def enterInst_error(self, ctx:pcodeParser.Inst_errorContext):
        pass

    # Exit a parse tree produced by pcodeParser#inst_error.
    def exitInst_error(self, ctx:pcodeParser.Inst_errorContext):
        pass


    # Enter a parse tree produced by pcodeParser#error.
    def enterError(self, ctx:pcodeParser.ErrorContext):
        pass

    # Exit a parse tree produced by pcodeParser#error.
    def exitError(self, ctx:pcodeParser.ErrorContext):
        pass



del pcodeParser


from dataclasses import dataclass
from enum import Enum
import inspect
import logging
import time
from typing import Any, Callable, Sequence

from openpectus.engine.hardware import HardwareLayerBase, HardwareLayerException, Register, RegisterDirection
from openpectus.engine.models import ConnectionStatusEnum
from openpectus.lang.exec.tags import Tag


logger = logging.getLogger(__name__)


class ErrorRecoveryState(Enum):
    Disconnected = 0
    """ No connection attempt yet """
    OK = 1
    """ No problem with the hardware connection """
    Issue = 2
    """ There may be an issue. A few errors detected. Trying to recover by waiting for success reads/writes
    while masking errors."""
    Reconnect = 3
    """ There is a connection an issue. Trying to recover by reconnecting while still masking errors."""
    Error = 4
    """ Connection is lost and reconnection has not been successful. Trying to recover by reconnecting.
    Errors are no longer masked."""


@dataclass
class ErrorRecoveryConfig():
    reconnect_timeout_seconds = 10
    """ Number of seconds in state Issue before changing to state Reconnect """
    error_timeout_seconds = 5*60*60
    """ Number of seconds in state Reconnect before changing to state Error """


class ErrorRecoveryDecorator(HardwareLayerBase):
    """ Implements error recovery as a decorator around concrete hardware, without coupling it to Engine.

    Error recovery has the 5 states defined in `ErrorRecoveryState`. It is configured using the time thresholds
    defined in `ErrorRecoveryConfig`.

    Once connected, state is `OK`. This state is kept until a read or write error occurs (`read()`, `write()` or
    one of the batch variants raise `HardwareLayerException`). Such an error causes state to transition to `Issue`.
    In this state, if a success read/write occurs, state transitions back to `OK`. If no success read/write occurs
    within a duration of `issue_timeout_seconds`, state is set to `Reconnect`.

    In state `Reconnect`, reconnect attempts are started. If successful, state is set to `OK`. If not successful
    within a duration of `error_timeout_seconds`, state is set to `Error`.

    While in states `Issue` and `Reconnect`, any read and write errors are masked by returning last-known-good values for
    reads and caching values for writes. This means that the Engine (and the user) will not notice the connection loss.

    The one exception to this is uod commands. We have to assume that a uod command cannot execute correctly
    without hardware connection so we have to fail uod commands. (A possible improvement would be to require uod
    commands to consider the connection and fail with predefined exception types).

    In state `Error` reconnect attempts continue but errors are no longer masked. The  `Connection Status` tag is set to
    'Disconnected'. If reconnect is successful, state is set to `OK` and `Connection Status` is set to `Connected`.

    The consequence of no longer masking errors, is that the Engine will enter the "paused on error" state where the
    user can decide whether to continue or not.

    Note: It is unlikely but possible that errors occur so soon after successful connection that no value is yet available
    as last-known-good. In this case, a read returns a None value is returned and the error is logged.

    Reconnect note:
    The default implementation uses the tick() method to detect and execute reconnect. If this takes too long
    and hurts engine timing, the hardware should instead implement its reconnect via threading.
    """

    def __str__(self) -> str:
        return f"ErrorRecoveryDecorator(state={self.state},decorated={type(self.decorated).__name__})"


    def __init__(self, decorated: HardwareLayerBase, config: ErrorRecoveryConfig, connection_status_tag: Tag) -> None:
        super().__init__()
        self.decorated = decorated
        self.config = config
        self.connection_status_tag = connection_status_tag
        self.connect_error_callback: Callable[[Exception], None] | None = None
        self.error_callback: Callable[[], None] | None = None
        self.reconnect_callback: Callable[[], None] | None = None
        self.reconnecting_callback: Callable[[], None] | None = None
        self.reconnected_callback: Callable[[], None] | None = None

        self.last_known_good_reads: dict[str, Any] = {}
        self.pending_writes: dict[Register, Any] = {}

        self.state: ErrorRecoveryState = ErrorRecoveryState.Disconnected

        self.last_success_read_write = time.time()
        self.last_success_connect = time.time()
        self.last_state_reconnect_time = time.time()
        """ The time of the last transition to state Reconnect"""

        self.reconnect_count = 0
        self.reconnect_tick = -1
        # Hardcoded ticks for exponential backoff. Multiples of the largest value are implicitly included
        self.reconnect_backoff_ticks = [5, 20, 100, 300, 1200, 18000]

        # support initialization with connected hwl, which is the default case
        if decorated.is_connected:
            logger.debug("Initializing recovery with state OK from connected hardware")
            self.state = ErrorRecoveryState.OK
            self.on_ok()
        else:
            logger.debug("Initializing recovery from disconnected hardware")

        try:
            self._setup_decorated_method_forwards()
        except Exception:
            logger.error("Error setting up decorated hardware method forwards", exc_info=True)

    def get_recovery_state(self) -> ErrorRecoveryState:
        return self.state

    @property
    def is_connected(self) -> bool:
        """ Returns a value indicating whether there is an active connection to the hardware."""
        return self.state in [ErrorRecoveryState.OK, ErrorRecoveryState.Issue]

    @property
    def registers(self):
        return self.decorated.registers

    def success_read_write(self):
        now = time.time()
        logger.debug(f"RW success, state: {self.state}")
        self.last_success_read_write = now
        if self.state == ErrorRecoveryState.Issue:
            logger.debug(f"RW success, state transition: {ErrorRecoveryState.Issue} -> {ErrorRecoveryState.OK}")
            self.state = ErrorRecoveryState.OK
            self.on_ok()

    def error_read_write(self):
        now = time.time()
        logger.debug(f"RW error, state: {self.state}")
        if self.state == ErrorRecoveryState.OK:
            logger.debug(f"RW error, state transition: {ErrorRecoveryState.OK} -> {ErrorRecoveryState.Issue}")
            self.state = ErrorRecoveryState.Issue
            self.on_issue()
        elif self.state == ErrorRecoveryState.Issue:
            if self.last_success_read_write + self.config.reconnect_timeout_seconds < now:
                logger.debug(f"RW error, state transition: {ErrorRecoveryState.Issue} -> {ErrorRecoveryState.Reconnect}")
                self.state = ErrorRecoveryState.Reconnect
                self.on_reconnect()
        elif self.state == ErrorRecoveryState.Reconnect:
            if self.last_state_reconnect_time + self.config.error_timeout_seconds < now:
                logger.debug(f"RW error, state transition: {ErrorRecoveryState.Reconnect} -> {ErrorRecoveryState.Error}")
                self.state = ErrorRecoveryState.Error
                self.on_error()

    def _is_backoff_tick(self, tick: int) -> bool:
        if tick in self.reconnect_backoff_ticks:
            return True
        last = self.reconnect_backoff_ticks[-1]
        if tick > last and tick % last == 0:
            return True
        return False

    def tick(self):
        if self.state in [ErrorRecoveryState.Reconnect, ErrorRecoveryState.Error]:
            self.reconnect_tick += 1
            if self._is_backoff_tick(self.reconnect_tick):
                logger.info("Reconnecting")
                self.on_reconnecting()
                try:
                    self.reconnect()
                    self.reconnect_tick = -1
                    self.state = ErrorRecoveryState.OK
                    self.on_reconnected()
                except Exception:
                    logger.error("Reconnect failed", exc_info=True)

    def _update_connection_status(self):
        if self.state in [ErrorRecoveryState.Disconnected, ErrorRecoveryState.Error]:
            value = ConnectionStatusEnum.Disconnected
        else:
            value = ConnectionStatusEnum.Connected
        logger.debug(f"Setting connection status to: {value} based on state: {self.state}")
        self.connection_status_tag.set_value(str(value), time.time())

    # Callbacks

    def on_ok(self):
        logger.debug("on_ok fired")
        self.decorated._is_connected = True
        self._update_connection_status()

    def on_issue(self):
        logger.debug("on_issue fired")
        self._update_connection_status()

    def on_error(self):
        logger.debug("on_error fired")
        self._update_connection_status()
        self.decorated._is_connected = False
        if self.error_callback is not None:
            self.error_callback()

    def on_reconnect(self):
        logger.debug("on_reconnect fired")
        self._update_connection_status()
        self.last_state_reconnect_time = time.time()

    def on_reconnecting(self):
        logger.debug("on_reconnecting fired")
        self._update_connection_status()
        if self.reconnecting_callback is not None:
            self.reconnecting_callback()

    def on_reconnected(self):
        logger.debug("on_reconnected fired")
        self._update_connection_status()
        self.decorated._is_connected = True
        if self.reconnected_callback is not None:
            self.reconnected_callback()

    # Decorator impl

    def read(self, r: Register) -> Any:
        if RegisterDirection.Read not in r.direction:
            raise KeyError(f"Cannot read from register {r.name} which has direction {r.direction}")

        if self.state in [ErrorRecoveryState.Disconnected, ErrorRecoveryState.Error]:
            raise HardwareLayerException(f"Cannot read in state {self.state}")

        if self.state == ErrorRecoveryState.Reconnect:
            # in this state we are reconnecting. we do not want a possible success read to interfere
            # with reconnection. But we do want a possible transition to Error so we note an error.
            self.error_read_write()
            if r.name in self.last_known_good_reads.keys():
                return self.last_known_good_reads[r.name]
            else:
                logger.error(f"No known good value available for read of register {r.name}")
                return None

        try:
            value = self.decorated.read(r)
            self.last_known_good_reads[r.name] = value
            self.success_read_write()
            return value
        except HardwareLayerException:
            self.error_read_write()
            return self._get_last_known_good_values([r])[0]

    def read_batch(self, registers: Sequence[Register]) -> list[Any]:
        for r in registers:
            if RegisterDirection.Read not in r.direction:
                raise KeyError(f"Cannot read from register {r.name} which has direction {r.direction}")

        if self.state in [ErrorRecoveryState.Disconnected, ErrorRecoveryState.Error]:
            raise HardwareLayerException(f"Cannot read in state {self.state}")

        if self.state == ErrorRecoveryState.Reconnect:
            self.error_read_write()
            return self._get_last_known_good_values(registers)

        try:
            values = self.decorated.read_batch(registers)
            for value, r in zip(values, registers):
                self.last_known_good_reads[r.name] = value
            self.success_read_write()
            return values
        except HardwareLayerException:
            self.error_read_write()
            return self._get_last_known_good_values(registers)

    def _get_last_known_good_values(self, registers: Sequence[Register]) -> list[Any]:
        values = []
        for r in registers:
            if r.name in self.last_known_good_reads.keys():
                values.append(self.last_known_good_reads[r.name])
            else:
                logger.error(f"Masked read failed. Register {r.name} did not have a last-known-good value")
                values.append(None)
        return values

    def write(self, value: Any, r: Register) -> None:
        if RegisterDirection.Write not in r.direction:
            raise KeyError(f"Cannot write to register {r.name} which has direction {r.direction}")

        if self.state in [ErrorRecoveryState.Disconnected, ErrorRecoveryState.Error]:
            raise HardwareLayerException(f"Cannot write in state {self.state}")

        if self.state == ErrorRecoveryState.Reconnect:
            self.error_read_write()
            self.pending_writes[r] = value
            return

        try:
            self.decorated.write(value, r)
            self.success_read_write()
            self._write_pending_values(except_names=[r.name])
        except HardwareLayerException:
            self.error_read_write()
            if self.state == ErrorRecoveryState.Error:
                return
            self.pending_writes[r] = value


    def write_batch(self, values: Sequence[Any], registers: Sequence[Register]):
        for r in registers:
            if RegisterDirection.Write not in r.direction:
                raise KeyError(f"Cannot write to register {r.name} which has direction {r.direction}")

        if self.state in [ErrorRecoveryState.Disconnected, ErrorRecoveryState.Error]:
            raise HardwareLayerException(f"Cannot write in state {self.state}")

        if self.state == ErrorRecoveryState.Reconnect:
            self.error_read_write()
            for value, r in zip(values, registers):
                self.pending_writes[r] = value
            return

        try:
            self.decorated.write_batch(values, registers)
            self.success_read_write()
            self._write_pending_values(except_names=[r.name for r in registers])
        except HardwareLayerException:
            self.error_read_write()
            if self.state == ErrorRecoveryState.Error:
                return
            for value, r in zip(values, registers):
                self.pending_writes[r] = value

    def connect(self):
        try:
            self.decorated.connect()

            if self.state == ErrorRecoveryState.Disconnected:
                self.state = ErrorRecoveryState.OK
                self.on_ok()

            self.connection_state = ConnectionStatusEnum.Connected
            self.last_success_connect = time.time()
        except HardwareLayerException:
            logger.error("Decorated connect failed", exc_info=True)
            raise

    def disconnect(self):
        return self.decorated.disconnect()

    def _write_pending_values(self, except_names: list[str] = []):
        if self.state == ErrorRecoveryState.OK and len(self.pending_writes) > 0:
            pending_items = list(self.pending_writes.items())
            for register, value in pending_items:
                if register.name in except_names:
                    continue
                try:
                    self.decorated.write(value, register)
                    del self.pending_writes[register]
                except HardwareLayerException:
                    pass  # Better luck next time. We just had a success write so chances are good

    def _setup_decorated_method_forwards(self):
        """ Set up method forwards to avoid the decorator breaking commands that are implemented as methods
            on the concrete hardware
        """
        def callable_publics_predicate(member):
            if not inspect.ismethod(member) and not inspect.isfunction(member):
                return False
            if (name := getattr(member, "__name__", None)) is not None:
                if name.startswith("__"):
                    return False
            return True

        decorated_members = inspect.getmembers_static(self.decorated, callable_publics_predicate)
        self_member_names = [member[0] for member in inspect.getmembers_static(self, callable_publics_predicate)]
        for member in decorated_members:
            name: str = member[0]
            if name not in self_member_names:
                logger.info(f"Forwarding method {name} from hardware class {type(self.decorated).__name__} to decorator")
                func = member[1]

                # define local method that forwards calls to func using self.decorated as its self
                def caller(*args, **kvargs):
                    return func(self.decorated, *args, **kvargs)
                setattr(self, name, caller)



from openpectus.aggregator.aggregator import Aggregator
from openpectus.aggregator.frontend_publisher import FrontendPublisher
from openpectus.protocol.aggregator_dispatcher import AggregatorDispatcher

_server: Aggregator | None = None


def get_aggregator() -> Aggregator:
    if _server is None:
        raise Exception("DI configuration error. Aggregator has not been initialized")
    return _server


def _create_aggregator(dispatcher: AggregatorDispatcher, publisher: FrontendPublisher, secret: str) -> Aggregator:
    global _server
    if _server is not None:
        return _server
    else:
        _server = Aggregator(dispatcher, publisher, secret)
        # print("GLOBAL: Creating aggregator server")
        return _server



import unittest

import pint
from pint import UndefinedUnitError, UnitRegistry, Quantity, Unit, DimensionalityError
from pint.util import UnitsContainer

ureg = UnitRegistry()
Q_ = Quantity
U_ = Unit


class PintTest(unittest.TestCase):

    def test_basics_quantity(self):
        distance = 24.0 * ureg.meter
        assert isinstance(distance, Quantity)
        self.assertIsInstance(distance, Quantity)
        self.assertEqual("24.0 meter", str(distance))

        self.assertEqual(24.0, distance.magnitude)

        self.assertIsInstance(distance.units, Unit)
        self.assertEqual("meter", str(distance.units))

        self.assertIsInstance(distance.dimensionality, UnitsContainer)
        self.assertEqual("[length]", distance.dimensionality)
        self.assertFalse(distance.dimensionless)

        velocity = 7 * ureg.meter / 1 * ureg.second
        self.assertIsInstance(velocity.dimensionality, UnitsContainer)
        self.assertEqual("[length] * [time]", velocity.dimensionality)

        weight = 2 * ureg.kg
        assert isinstance(weight, Quantity)
        self.assertEqual("kilogram", str(weight.units))
        self.assertEqual("[mass]", weight.dimensionality)

        volume = 3 * ureg.liter  # hmm, no sign of 'volume' anywhere in the api
        assert isinstance(volume, Quantity)
        self.assertEqual("liter", str(volume.units))
        self.assertEqual("[length] * [length] * [length]", volume.dimensionality)

        time = Q_("1 min")  # Note: need 'min' as 'm' is meter
        self.assertEqual("minute", str(time.units))
        self.assertEqual("[time]", time.dimensionality)

        length = Q_("1 m")
        self.assertEqual("meter", str(length.units))
        self.assertEqual("[length]", length.dimensionality)

    def test_basics_unit(self):
        u = U_("s")
        self.assertFalse(u.dimensionless)
        self.assertEqual("second", str(u))
        self.assertEqual("[time]", str(u.dimensionality))
        # print("{!r}".format(u))

        q = 5 * u
        # print("{!r}".format(q))

    def test_dimensionless(self):
        count = Q_(3)
        self.assertTrue(count.dimensionless)
        self.assertEqual("<Quantity(3, 'dimensionless')>", "{!r}".format(count))

        count = Q_(3, None)
        self.assertTrue(count.dimensionless)
        self.assertEqual("<Quantity(3, 'dimensionless')>", "{!r}".format(count))

        u1 = U_("")  # must use empty string, not None or no argument
        self.assertTrue(u1.dimensionless)
        self.assertEqual("<Unit('dimensionless')>", u1.__repr__())

    def test_dimensionless_percentage(self):
        count = Q_(3, "%")
        self.assertTrue(count.dimensionless)
        self.assertTrue(count.unitless)
        self.assertEqual("<Quantity(3, 'percent')>", "{!r}".format(count))
        self.assertEqual(count.units.__repr__(), "<Unit('percent')>")

    def test_pressure(self):
        val = Q_(1, "bar")
        exp_dimensionality = "[mass] / [length] / [time] ** 2"
        self.assertEqual(exp_dimensionality, str(val.dimensionality))

        compatibles = ureg.get_compatible_units(exp_dimensionality)  # type: ignore
        self.assertIn('bar', [str(c) for c in compatibles])

        # Note: 'Pa' is not in compatibles but 'pascal' is
        self.assertNotIn('Pa', [str(c) for c in compatibles])
        self.assertIn('pascal', [str(c) for c in compatibles])

        # but Pa is a valid unit
        val2 = Q_(1, "Pa")
        self.assertEqual(exp_dimensionality, str(val2.dimensionality))

    def test_flow(self):
        val = Q_(1, "l/h")
        exp_dimensionality = "[length] ** 3 / [time]"
        self.assertEqual(exp_dimensionality, str(val.dimensionality))

    def test_mass_flow_rate(self):
        val = Q_(1, "kg/h")
        exp_dimensionality = "[mass] / [time]"
        self.assertEqual(exp_dimensionality, str(val.dimensionality))

    def test_frequency(self):
        val = Q_(1, "Hz")
        exp_dimensionality = "1 / [time]"
        self.assertEqual(exp_dimensionality, str(val.dimensionality))

    def test_absorbance(self):
        val = Q_(1, "mS/cm")
        exp_dimensionality = "[current] ** 2 * [time] ** 3 / [mass] / [length] ** 3"
        self.assertEqual(exp_dimensionality, str(val.dimensionality))

    def test_area(self):
        val_m2 = Q_(1, "m**2")
        unit_cm2 = Unit("cm**2")
        val_cm2 = val_m2.to(unit_cm2)
        self.assertAlmostEqual(10000, val_cm2.magnitude, delta=0.001)
        unit_dm2 = Unit("dm**2")
        val_dm2 = val_m2.to(unit_dm2)
        self.assertAlmostEqual(100, val_dm2.magnitude, delta=0.001)

    def test_formatting(self):
        weight = 2 * ureg.kg
        s = 'The magnitude is {0.magnitude} with units {0.units}'.format(weight)
        self.assertEqual("The magnitude is 2 with units kilogram", s)
        repr = 'The representation is {!r}'.format(weight)
        self.assertEqual("The representation is <Quantity(2, 'kilogram')>", repr)
        # or just use __repr__()
        self.assertEqual("<Quantity(2, 'kilogram')>", weight.__repr__())

    def test_conversion(self):
        distance_km = 3.2 * ureg.kilometers
        assert isinstance(distance_km, Quantity)
        distance_m = distance_km.to(ureg.meter)
        self.assertEqual(3200, distance_m.magnitude)

    def test_conversion_invalid(self):
        distance_km = 3.2 * ureg.kilometers
        assert isinstance(distance_km, Quantity)
        with self.assertRaises(DimensionalityError) as err:
            _ = distance_km.to(ureg.second)
        self.assertEqual("Cannot convert from 'kilometer' ([length]) to 'second' ([time])", str(err.exception))

    def test_compatibility_quantity(self):
        dist = Q_("310m")
        weight = Q_("3kg")
        temp = Q_(3, ureg.degC)
        time = Q_("5 sec")

        # NOTE: don't use ureg.is_compatible_with:
        # self.assertFalse(ureg.is_compatible_with(dist1, weight))  # this fails
        # Instead, use the Quantity.is_compatible_with() method:
        self.assertFalse(dist.is_compatible_with(weight))
        self.assertFalse(dist.is_compatible_with(temp))
        self.assertFalse(time.is_compatible_with(temp))

        # or use this for testing compability then
        with self.assertRaises(DimensionalityError):
            _ = dist.to(ureg.second)
        with self.assertRaises(DimensionalityError):
            _ = dist.to(ureg.mass)

        # or maybe just check dimensionality with the check() method
        self.assertTrue(dist.check('[length]'))
        self.assertTrue(Q_("3kg").check('[mass]'))

    def test_compatibility_unit(self):
        second = pint.Unit("sec")  # the default constructor
        hour = U_("h")  # or the short hand
        kg = U_("kg")

        self.assertTrue(second.is_compatible_with(hour))
        self.assertFalse(second.is_compatible_with(kg))

    @unittest.skip("Not sure we need this. Requres the Babel lib")
    def test_locale(self):
        # print("locale:", ureg.fmt_locale)  # is None by default
        # set_fmt_locale: None: (do not translate), 'sys' (detect the system locale) or a locale id string.
        ureg.set_fmt_locale('sys')
        distance = 24.0 * ureg.meter
        self.assertEqual("24.0 meter", str(distance))

    def test_dimensions(self):
        # temperature = 9.5 * ureg.degC  # fails, because degC is not a multiplicative unit,
        # see https://pint.readthedocs.io/en/0.10.1/nonmult.html

        # this works. but beware of deltas. temperature calculus is tricky
        # alternatively, use ureg = UnitRegistry(autoconvert_offset_to_baseunit = True)
        # to perform automatic conversions
        temperature = Q_(25.4, ureg.degC)
        self.assertEqual("25.4 degree_Celsius", str(temperature))

    def test_higher_dimensionality(self):
        flow = Q_("10 L/h")
        self.assertIsInstance(flow, Quantity)
        self.assertEqual("<Quantity(10.0, 'liter / hour')>", "{!r}".format(flow))
        self.assertEqual("liter / hour", str(flow.units))

        flow = Q_("10 L/d")
        self.assertIsInstance(flow, Quantity)
        self.assertEqual("<Quantity(10.0, 'liter / day')>", "{!r}".format(flow))

        flow = Q_("10 L/min")
        self.assertIsInstance(flow, Quantity)
        self.assertEqual("<Quantity(10.0, 'liter / minute')>", "{!r}".format(flow))

    def test_string_parsing(self):
        distance = Q_("24cm")
        self.assertIsInstance(distance, Quantity)
        self.assertEqual("24 centimeter", str(distance))

        distance = 2.54 * ureg('nm')
        self.assertIsInstance(distance, Quantity)
        self.assertEqual("2.54 nanometer", str(distance))

        volume = Q_("3 L")
        self.assertEqual("<Quantity(3, 'liter')>", "{!r}".format(volume))

        count = Q_(7)
        self.assertEqual("<Quantity(7, 'dimensionless')>", "{!r}".format(count))

    def test_comparison(self):
        a = Q_("10 cm")
        b = Q_("20 cm")
        c = Q_("0.2 m")
        self.assertTrue(a < b)
        self.assertTrue(a < c)
        self.assertEqual(b, c)

    def test_test_for_unknown_unit(self):
        with self.assertRaises(UndefinedUnitError):
            _ = Q_("7.0 CV")

    @unittest.skip(reason="Not a test. Merely documenttion of ureg.get_compatible_units")
    def test_list_compatible_units(self):
        a = Q_("10 cm")
        result = ureg.get_compatible_units(a.units)  # type: ignore
        print(result)
        # Returns lots of useless values like parsec, nautical mile, angstrom_star, light_year and such
        # and a few useful ones, like meter. Need to rool our own or customize pint defaults
        # see https://github.com/hgrecco/pint/blob/master/pint/default_en.txt

    # def test_customizing_pint_registry(self):
    #     ureg = UnitRegistry(cache_folder=":auto:")
    #     print("ureg.cache_folder", ureg.cache_folder)


if __name__ == "__main__":
    unittest.main()



class ProtocolException(Exception):
    """ Exception for errors related to the Aggregator-Engine protocol """
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class ProtocolDeserializationException(ProtocolException):
    """ Raised on deserialization errors. If this error occurs, it indicates an implementation error. """
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class ProtocolNetworkException(Exception):
    """ Raised on errors related to the physical network connection """
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



import time
import unittest

from openpectus.lang.exec.timer import OneThreadTimer


@unittest.skip("Slow")
class TimerTest(unittest.TestCase):

    def test_one_thread_timer(self):
        def tick(tick_time: float, increment_time: float):
            print('tick: {:.4f}'.format(tick_time))
            time.sleep(0.3)  # simulate work

        timer = OneThreadTimer(1.0, tick)
        timer.start()
        print('timer.start() returned')

        time.sleep(5.5)
        timer.stop()
        print('timer.stopped')

    def test_one_thread_timer_drift(self):
        def tick(tick_time: float, increment_time: float):
            print("tick_time", tick_time)
            time.sleep(0.07)  # simulate safe work amount
            # time.sleep(0.09)  # simulate unsafe work amount - will print warnings

        interval = .1
        timer = OneThreadTimer(interval, tick)
        timer.start()
        print('timer.start() returned')

        time.sleep(3)
        timer.stop()
        print('timer.stopped')

    def test_one_thread_timer_can_restart(self):
        def tick(tick_time: float, increment_time: float):
            print("tick_time", tick_time)
            time.sleep(0.8 * interval)  # simulate work

        interval = 0.3
        timer = OneThreadTimer(interval, tick)
        timer.start()
        print('timer.start() returned')

        time.sleep(5 * interval + .1)
        timer.stop()
        print('timer stopped')

        time.sleep(3 * interval + .1)

        timer.start()
        print('timer restarted')

        time.sleep(5 * interval + .1)
        timer.stop()
        print('timer stopped again')


if __name__ == "__main__":
    unittest.main()



from enum import StrEnum, auto

from fastapi.routing import APIRouter
from fastapi_websocket_pubsub import PubSubEndpoint
from fastapi_websocket_pubsub.event_notifier import TopicList
from fastapi_websocket_pubsub.rpc_event_methods import RpcEventServerMethods


class PubSubTopic(StrEnum):
    RUN_LOG = auto()
    METHOD = auto()
    CONTROL_STATE = auto()
    ERROR_LOG = auto()
    PROCESS_UNITS = auto()


class FrontendPublisher:
    class MethodsWithUnsubscribe(RpcEventServerMethods):
        """ We need to add our own unsubscribe method to support this for the frontend client.
            If/when I get this added to the library itself, we can remove this """

        async def unsubscribe(self, topics: TopicList = []) -> bool:
            channel_id = await self._get_channel_id_()
            await self.event_notifier.unsubscribe(channel_id, topics)
            return True

    def __init__(self):
        self.router = APIRouter(tags=["frontend_pubsub"], prefix='/api')
        self.router.add_api_route('/expose-pubsub-topics', endpoint=self.expose_pubsub_topics, methods=['POST'])
        self.router.add_api_route('/trigger-publish-msw', endpoint=self.trigger_publish_msw, methods=['POST'])
        self.pubsub_endpoint = PubSubEndpoint(methods_class=FrontendPublisher.MethodsWithUnsubscribe)
        self.pubsub_endpoint.register_route(self.router, path="/frontend-pubsub")

    async def publish_process_units_changed(self):
        await self.pubsub_endpoint.publish(PubSubTopic.PROCESS_UNITS)

    async def publish_run_log_changed(self, unitId: str):
        await self.pubsub_endpoint.publish(f'{unitId}/{PubSubTopic.RUN_LOG}')

    async def publish_method_changed(self, unitId: str):
        await self.pubsub_endpoint.publish(f'{unitId}/{PubSubTopic.METHOD}')

    async def publish_method_state_changed(self, unitId: str):
        # for now we use PubSubTopic.METHOD for both Method and MethodState changes
        await self.pubsub_endpoint.publish(f'{unitId}/{PubSubTopic.METHOD}')

    async def publish_control_state_changed(self, unitId: str):
        await self.pubsub_endpoint.publish(f'{unitId}/{PubSubTopic.CONTROL_STATE}')

    async def publish_error_log_changed(self, unitId: str):
        await self.pubsub_endpoint.publish(f'{unitId}/{PubSubTopic.ERROR_LOG}')

    def expose_pubsub_topics(self, topic: PubSubTopic):
        """ This endpoint is just for exposing the topic enum to frontend via autogeneration """
        pass

    async def trigger_publish_msw(self):
        """ Publish to all topics that start with 'MSW_' """
        topics_subscribed_to = list(self.pubsub_endpoint.notifier._topics.keys())
        msw_topics_subscribed_to = [topic for topic in topics_subscribed_to if topic.startswith('MSW_')]
        await self.pubsub_endpoint.publish(msw_topics_subscribed_to)



import math
from time import time
from typing import Any

from openpectus.engine.uod_builder_api import (
    UnitOperationDefinitionBase, UodBuilder, UodCommand,
    as_float,
    tags,
    PlotConfiguration, SubPlot, PlotAxis, PlotColorRegion,
    RegexNumber, RegexCategorical,
    HardwareLayerBase, Register, RegisterDirection  # special import required by custom hardware, DemoHardware
)
from openpectus.lang.exec.tags import format_time_as_clock


def create() -> UnitOperationDefinitionBase:
    builder = UodBuilder()
    # use this logger to send warnings/errors to the frontend error log
    # logger = builder.get_logger()

    # this signature is usable for the default argument parser
    # def reset(cmd: UodCommand, **kvargs):

    # this signature is valid for arg_parse_fn=None
    def reset(cmd: UodCommand):
        """ # Resets the demo unit
        Reset
        """
        count = cmd.get_iteration_count()
        max_ticks = 100
        if count == 0:
            cmd.context.tags.get("Reset").set_value("Reset", time())
            cmd.context.hwl.reset_FT01()  # type: ignore
        elif count > max_ticks:
            cmd.context.tags.get("Reset").set_value("N/A", time())
            cmd.set_complete()
        else:
            progress = count/max_ticks
            cmd.set_progress(progress)

    def test_int(cmd: UodCommand, value):
        """# Runs test command
        TestInt: 5
        """
        print("test_cmd executing with arg: " + value)
        fval = as_float(value)
        if fval is None:
            # raising ValueError will display the error to the user
            raise ValueError(f"value '{value}' is not a number")
        cmd.context.tags.get("TestInt").set_value(fval, time())
        cmd.set_complete()

    def cmd_regex(cmd: UodCommand, number: str, number_unit: str | None = None):
        """# Set the CmdWithRegexArgs tag value
        CmdWithRegexArgs: 5 dm2
        """
        # optional arg is ok when regex's named groups do not include it
        print(f"cmd_regex executing with number: {number} and number_unit: {number_unit}")
        assert number_unit is not None
        cmd.context.tags.get("CmdWithRegexArgs").set_value_and_unit(float(number), number_unit, time())
        cmd.set_complete()

    def cmd_regex_categorical(cmd: UodCommand, option: str):
        """# Set the TestCategorical tag value
        TestCategorical: A
        """
        print(f"cmd_regex_categorical executing with option: {option}")
        cmd.context.tags.get("TestCategorical").set_value(option, time())
        cmd.set_complete()

    def test_percentage(cmd: UodCommand, number, number_unit):
        # Note: When using RegexNumber, the number argument is still a
        # string but it will always support conversion to float
        number = float(number)
        cmd.context.tags.get("TestPercentage").set_value(number, time())
        cmd.set_complete()

    def set_category(cmd: UodCommand, value: str):
        """# Set the category tag.

        # This is just an example and it doesn't do anything because the
        # tag is controlled and constantly updated by the demo hardware.
        Category: Falling"""
        cmd.context.tags["Category"].set_value(value, time())
        cmd.set_complete()

    def get_plot_configuration() -> PlotConfiguration:
        return PlotConfiguration(
            color_regions=[
                PlotColorRegion(
                    process_value_name='Category',
                    value_color_map={'Rising': '#66ff6633', 'Falling': '#6666ff33'})
            ],
            sub_plots=[SubPlot(
                axes=[PlotAxis(
                    label='FT01',
                    process_value_names=['FT01'],
                    y_max=20,
                    y_min=0,
                    color='#ff0000',
                ), PlotAxis(
                    label='FT02',
                    process_value_names=['FT02'],
                    y_max=20,
                    y_min=0,
                    color='#0000ff',
                )],
                ratio=1
            )],
            process_value_names_to_annotate=['Category'],
            x_axis_process_value_names=['Time']
        )

    return (
        builder
        .with_instrument("DemoUod")
        .with_author("Demo Author", "demo@openpectus.org")
        .with_filename(__file__)
        # .with_required_roles(['RoleTest'])
        .with_hardware(DemoHardware())
        .with_location("Demo location")
        .with_hardware_register("FT01", RegisterDirection.Read, path='Objects;2:System;2:FT01')
        .with_hardware_register("FT02", RegisterDirection.Read, path='Objects;2:System;2:FT02')
        .with_hardware_register("Category", RegisterDirection.Read, path='Objects;2:System;2:Category')
        .with_hardware_register("Time", RegisterDirection.Read)
        .with_hardware_register("Reset", RegisterDirection.Both, path='Objects;2:System;2:RESET',
                                from_tag=lambda x: 1 if x == 'Reset' else 0,
                                to_tag=lambda x: "Reset" if x == 1 else "N/A")
        .with_tag(tags.ReadingTag("FT01", "L/h"))
        .with_tag(tags.ReadingTag("FT02", "L/h"))
        .with_tag(tags.SelectTag("Category", "Rising", unit=None, choices=["Rising", "Falling"]))
        .with_tag(tags.ReadingTag("Time", unit="s", format_fn=format_time_as_clock))
        .with_tag(tags.Tag("TestInt", value="42"))
        .with_tag(tags.Tag("TestFloat", value=9.87, unit="kg"))
        .with_tag(tags.Tag("TestString", value="test"))
        .with_tag(tags.Tag("TestCategorical", value=""))
        .with_tag(tags.SelectTag("Reset", value="N/A", unit=None, choices=['Reset', "N/A"]))
        .with_tag(tags.Tag("CmdWithRegexArgs", value=34.87, unit="dm2"))
        .with_tag(tags.Tag("TestPercentage", value=34.87, unit="%"))
        .with_command(name="Reset", exec_fn=reset, arg_parse_fn=None)
        .with_command(name="TestInt", exec_fn=test_int)
        .with_command(name="Category", exec_fn=set_category)
        .with_process_value(tag_name="Run Time")
        .with_process_value(tag_name="FT01")
        .with_process_value_entry(tag_name="TestInt")
        .with_process_value_entry(tag_name="TestFloat", execute_command_name="TestInt")
        .with_process_value_choice(tag_name="TestString", command_options={'A': 'Mark: A', 'B': 'Mark: B'})
        .with_process_value_choice(tag_name="Category",
                                   command_options={"Rising": "Category: Rising", "Falling": "Category: Falling"})
        .with_process_value(tag_name="FT02")
        .with_process_value(tag_name="Time")
        .with_process_value_choice(tag_name="Reset", command_options={'Reset': 'Reset'})
        .with_process_value(tag_name="System State")
        .with_process_value(tag_name="CmdWithRegexArgs")
        .with_command_regex_arguments(
            name="CmdWithRegexArgs",
            arg_parse_regex=RegexNumber(units=['m2', 'dm2']),
            exec_fn=cmd_regex)
        .with_command_regex_arguments(
            name="TestPercentage",
            arg_parse_regex=RegexNumber(units=['%']),
            exec_fn=test_percentage)
        .with_command_regex_arguments(
            name="TestCategorical",
            arg_parse_regex=RegexCategorical(exclusive_options=["A", "B"],
                                             additive_options=["1", "2", "3"]),
            exec_fn=cmd_regex_categorical)
        .with_process_value_entry(tag_name="TestPercentage")

        .with_plot_configuration(get_plot_configuration())
        .build()
    )


class DemoHardware(HardwareLayerBase):
    """ Simulation hardware. """

    def __init__(self) -> None:
        super().__init__()
        self.start_time: float = time()

    def reset_FT01(self):
        self.start_time = time()

    def read(self, r: Register) -> Any:
        if r.name == "FT01":
            elapsed_seconds = time() - self.start_time
            half_minutes = elapsed_seconds / 30
            return int(10 * (math.sin(half_minutes) + 1))
        elif r.name == "FT02":
            elapsed_seconds = time() - self.start_time
            half_minutes = elapsed_seconds / 30
            return 10 * (math.sin(half_minutes) + 1)
        elif r.name == "Time":
            return time() - self.start_time
        elif r.name == 'Category':
            elapsed_seconds = time() - self.start_time
            half_minutes = elapsed_seconds / 30
            return 'Rising' if math.cos(half_minutes) > 0 else 'Falling'

    def write(self, value: Any, r: Register):
        pass
        # if r.name == "Reset":
        #     return
        # print(f"DemoHardware: Wrote value '{value}' to register '{r.name}'")



from __future__ import annotations

import copy
import logging
from enum import StrEnum, auto
from typing import Dict
from uuid import UUID, uuid4

from openpectus.engine.commands import EngineCommand
from openpectus.lang.exec.tags import TagValueCollection
from openpectus.lang.exec.uod import UodCommand
from openpectus.lang.model.pprogram import (
    PAlarm, PBlank, PComment, PInjectedNode, PNode, PProgram, PErrorInstruction, PWatch
)

logger = logging.getLogger(__name__)


class RuntimeInfo():
    """ Provides a log of the actions that take place during interpretation.

    It consists of a list of records. Each record has a name and a list of
    states. The record state contain a state name, time and the tag values
    at the time the state was created.

    As interpretation progresses, the program AST nodes (subclasses of PNode)
    are visited and their semantics are executed. The records added are:

    - For all visited nodes, a record is added. The name of the record is
      provided by PNode.runlog_name if overwritten by the particular subclass.
      When a record is created, its exec_id is given a random UUID.

    - Instructions that are interpreted via an interrupt, reuse the record
      that was created when the node was first visited.

      Alarms too?

    - Engine commands are currently not contained in records.

    - Uod commands are executed via context.schedule_execution(). The interpreter
      does not know what happens after that. The engine updates the record as
      these commands are executed, cancelled etc. The record states are provided
      with the command instance as well as a command_exec_id that tracks a
      particular command instance. This means that one record may hold states
      for different invocations (and instances) of the same command.
    """

    def __init__(self) -> None:
        self._records: list[RuntimeRecord] = []
        self._record_index: Dict[UUID, int] = {}

    @property
    def records(self) -> list[RuntimeRecord]:
        return list(self._records)

    def get_runlog(self) -> RunLog:
        """ Transmogrify the runtime records into the RunLog shape that clients
        understand.

        Notably, it removes records with no name and unscrambles uod states
        such that each invocation gets its own runlog item. This allows the client
        to force or cancel a particular command instance.
        """
        runlog = RunLog()
        runlog.id = self._get_runlog_id() or ""
        for r in self.records:
            try:
                items = self._get_record_runlog_items(r)
                runlog.items.extend(items)
            except Exception:
                logger.error(f"Failed to create runlog item for record {r}", exc_info=True)
        runlog.items.sort(key=lambda item: item.start)
        return runlog

    def _get_runlog_id(self) -> str | None:
        for r in self.records:
            if isinstance(r.node, PProgram):
                return str(r.node.id)

    def _get_record_runlog_items(self, r: RuntimeRecord) -> list[RunLogItem]:
        if isinstance(r.node, (PProgram, PBlank, PComment)):
            return []
        if r.name is None:
            if isinstance(r.node, (PInjectedNode, PErrorInstruction)):
                return []
            node_name = str(r.node) if r.node is not None else "node is None"
            logger.error(f"Runtime record has empty name. node: {node_name}. Fix this error or add a rule exception.")
            return []
        if r.name == "Stop":
            return []

        items: list[RunLogItem] = []

        # Usually there is only a single start/complete state pair which is what is needed for a runlog item.
        # But, alas, alarms can be invoked any number of times for which all state information is placed in
        # the same runtime record. Therefore we need to expand these states into seperate run log items.
        # And because any node can be in a PAlarm body, this extends to all other nodes as well.
        split_states = self._split_states(r, include_prestart_states=False)
        # iterate the invocations
        for invocation_inx, invocation_states in enumerate(split_states):
            # is_before_first_invocation = False
            # has_more_invocations = invocation_inx < len(split_states) - 1
            item: RunLogItem | None = None
            command: EngineCommand | None = None
            # iterate the states of the invocation
            for inx, state in enumerate(invocation_states):
                is_start_state = inx == 0
                has_more_states = inx < len(invocation_states) - 1
                is_conclusive_state = state.state_name in [
                    RuntimeRecordStateEnum.Completed, RuntimeRecordStateEnum.Failed, RuntimeRecordStateEnum.Cancelled
                ]

                if is_start_state:
                    item = RunLogItem()
                    item.name = r.name
                    item.id = str(r.exec_id)  # Note is changed to command_exec_id later for uod commands
                    item.state = RunLogItemState.Started  # TODO possibly improve - could also be Waiting
                    item.start = state.state_time
                    item.start_values = state.values or TagValueCollection.empty()
                    item.cancellable = r.node.cancellable
                    item.cancelled = r.node.cancelled
                    item.forcible = r.node.forcible
                    item.forced = r.node.forced

                if state.state_name == RuntimeRecordStateEnum.Completed:
                    assert item is not None
                    item.state = RunLogItemState.Completed
                elif state.state_name == RuntimeRecordStateEnum.Failed:
                    assert item is not None
                    item.state = RunLogItemState.Failed
                elif state.state_name == RuntimeRecordStateEnum.Cancelled:
                    assert item is not None
                    item.state = RunLogItemState.Cancelled
                    item.cancelled = True
                elif state.state_name == RuntimeRecordStateEnum.Forced:
                    assert item is not None
                    item.state = RunLogItemState.Forced
                    item.forced = True
                elif state.state_name == RuntimeRecordStateEnum.UodCommandSet:
                    assert item is not None
                    logger.debug(f"Updating uod item id from record id {item.id} to" +
                                 f"record state command_exec_id {state.command_exec_id}")
                    item.id = str(state.command_exec_id)
                    command = state.command
                elif state.state_name == RuntimeRecordStateEnum.InternalEngineCommandSet:
                    assert item is not None
                    logger.debug(f"Updating internal engine item id from record id {item.id} to" +
                                 f"record state command_exec_id {state.command_exec_id}")
                    item.id = str(state.command_exec_id)
                    command = state.command

                if not is_conclusive_state and item is not None:
                    item.cancellable = r.node.cancellable
                    item.forcible = r.node.forcible
                    if command is not None:
                        if isinstance(command, UodCommand):
                            item.cancellable = True  # PCommand.cancellable does not support uod commands
                        self._update_item_progress(item, command)

                if is_conclusive_state:
                    assert item is not None
                    item.end = state.state_time
                    item.end_values = state.values or TagValueCollection. empty()
                    item.cancellable = False
                    item.forcible = False

                if not has_more_states or is_conclusive_state:
                    if item is not None:
                        items.append(item)
                        item = None
                        command = None

        return items

    def _split_states(self, r: RuntimeRecord, include_prestart_states: bool) -> list[list[RuntimeRecordState]]:
        """ Split a record's states into those of distint invocations."""
        if len(r.states) == 0:
            return []

        split_states: list[list[RuntimeRecordState]] = []

        if r.has_state(RuntimeRecordStateEnum.UodCommandSet):
            # UodCommandSet signifies the start of a uod command invocation
            split_states = self._split_states_by_state_name(
                r.states, RuntimeRecordStateEnum.UodCommandSet, include_prestart_states)
        elif isinstance(r.node, (PAlarm, PWatch)):
            # AwaitingCondition signifies the start of a new invocation for alarm and watch nodes
            split_states = self._split_states_by_state_name(
                r.states, RuntimeRecordStateEnum.AwaitingCondition, include_prestart_states)
        else:
            split_states = self._split_states_by_state_name(
                r.states, RuntimeRecordStateEnum.Started, include_prestart_states)

        return split_states

    def _split_states_by_state_name(
            self,
            states: list[RuntimeRecordState],
            state_name: RuntimeRecordStateEnum,
            include_prestart_states: bool) -> list[list[RuntimeRecordState]]:
        if len(states) == 0:
            return []

        # we need ordered states. if we cannot assume this, we would need to sort them
        self._check_record_states_ordered(states)

        split_states: list[list[RuntimeRecordState]] = []
        cur_instance_states: list[RuntimeRecordState] = []
        start_state_indices = [i for i, st in enumerate(states) if st.state_name == state_name]
        for i, st in enumerate(states):
            if len(start_state_indices) > 0 and i < min(start_state_indices):  # before the first start state
                if include_prestart_states:
                    # these could be used to produce a meta/root item for the invocation items
                    cur_instance_states.append(st)
            elif i in start_state_indices:  # at a start state
                if len(cur_instance_states) > 0:
                    split_states.append(cur_instance_states)
                cur_instance_states = [st]
            else:  # after a start state
                cur_instance_states.append(st)
        if len(cur_instance_states) > 0:
            split_states.append(cur_instance_states)
        return split_states

    def _check_record_states_ordered(self, states: list[RuntimeRecordState]):
        # temporarily verify that states are always ordered
        # if this is consistently the case, we can maybe ensure it as an invariant and remove the check
        for i, state in enumerate(states):
            if i > 0:
                prev_state = states[i-1]
                try:
                    assert prev_state.state_tick <= state.state_tick, "State tick out of order"
                    assert prev_state.state_time <= state.state_time, "State time out of order"
                except Exception:
                    raise

    def _update_item_progress(self, item: RunLogItem, command: EngineCommand):
        progress = command.get_progress()
        if isinstance(progress, float):
            item.progress = progress
        elif isinstance(progress, bool) and progress:
            item.progress = 0.5
        else:
            item.progress = None
        # logger.info(f"Updating progress to {item.progress}")

    def get_exec_record(self, exec_id: UUID) -> RuntimeRecord | None:
        """ Return record for the given exec_id or None if not found. """
        index = self._record_index.get(exec_id)
        if index is not None:
            return self._records[index]

    def get_command_and_record(self, command_exec_id: UUID) -> tuple[EngineCommand, RuntimeRecord] | None:
        """ Return (command, record) pair for the given command_exec_id or None if (both) not found. """
        for record in reversed(self.records):
            for state in record.states:
                if state.command_exec_id is not None and state.command_exec_id == command_exec_id:
                    if state.command is not None:
                        return state.command, record

    def get_last_node_record_or_none(self, node: PNode) -> RuntimeRecord | None:
        for r in reversed(self._records):
            if r.node == node:
                return r

    def get_last_node_record(self, node: PNode) -> RuntimeRecord:
        record = self.get_last_node_record_or_none(node)
        if record is None:
            raise ValueError("Node has no records")
        return record

    def get_node_records(self, node: PNode) -> list[RuntimeRecord]:
        return [r for r in self._records if r.node.id == node.id]

    def _add_record(self, record: RuntimeRecord, exec_id: UUID):
        index = len(self._records)
        self._records.append(record)
        self._record_index[exec_id] = index
        assert self._records[index].exec_id == exec_id
        self._last_exec_id = exec_id

    def begin_visit(
            self,
            node: PNode,
            time: float, tick: int,
            start_values: TagValueCollection) -> RuntimeRecord:
        exec_id = uuid4()
        record = RuntimeRecord(node, exec_id=exec_id)
        record.visit_start_time = time
        record.visit_start_tick = tick
        record.start_values = start_values
        self._add_record(record, exec_id=exec_id)
        return record

    def find_instruction(
            self,
            instruction_name: str,
            start_index: int,
            instruction_state: RuntimeRecordStateEnum | None
            ) -> int | None:
        """ Find the first record with the given instruction name and state starting from start_index (incl).

        Return the index of the found record or None if record is not found.

        Use get_record_by_index to obtain the record of the index.
        """
        if start_index < 0:
            raise ValueError(f"start_index {start_index} is invalid")

        # create a copy to avoid race conditions cause by timer tick
        records_copy = self.records.copy()
        size = len(records_copy)
        if start_index > size - 1:
            return None

        for i in range(start_index, size):
            r = records_copy[i]
            if r.node.instruction_name == instruction_name:
                # logger.debug(f"Find record result: {i}, {instruction_name=}, {start_index=}")
                # logger.debug(self.get_as_table())
                if instruction_state is None:
                    return i
                elif r.has_state(instruction_state):
                    return i

    def get_record_by_index(self, index: int) -> RuntimeRecord | None:
        if index < 0:
            raise ValueError(f"index {index} is invalid")
        records = self.records.copy()
        if index > len(records) - 1:
            return None
        else:
            return copy.copy(records[index])

    def get_as_table(self, description: str = "") -> str:
        records = self.records.copy()
        lines = [f"Runtime records: {description}"]
        lines.append("line | start | end   | name                 | instruction name     | states")
        lines.append("-----|-------|-------|----------------------|----------------------|-------------------")
        for r in records:
            name = f"{str(r.name):<20}" if r.name is not None else f"{str(r.node):<20}"
            inst_name = f"{str(r.node.instruction_name):<20}" \
                if r.node.instruction_name is not None else "   -                  "
            line = f"{int(r.node.line):4d}" if r.node.line is not None else "   -"
            states = ", ".join([f"{st.state_name}: {st.state_tick}" for st in r.states])
            end = f"{r.visit_end_tick:5d}" if r.visit_end_tick != -1 else "    -"
            lines.append(f"{line}   {r.visit_start_tick:5d}   {end}   {name}   {inst_name}   {states}")
        lines.append("-----|-------|-------|----------------------|----------------------|-------------------")
        return "\n".join(lines)

    def print_as_table(self, description: str = ""):
        print(self.get_as_table(description))

class RuntimeRecord():
    def __init__(self, node: PNode, exec_id: UUID) -> None:
        self.exec_id: UUID = exec_id
        self.node = node
        self.name = node.runlog_name
        self.visit_start_time: float = -1.0
        self.visit_start_tick: int = -1
        self.visit_end_time: float = -1.0
        self.visit_end_tick: int = -1

        self.states: list[RuntimeRecordState] = []
        self.start_values: TagValueCollection | None = None
        self.end_values: TagValueCollection | None = None

    @staticmethod
    def null_record() -> RuntimeRecord:
        # Returns a Null Object value that can be used when a real value is not available
        return RuntimeRecord(PNode(None), exec_id=uuid4())

    def __repr__(self) -> str:
        return f"{self.name} | States: {', '.join([str(st) for st in self.states])}"

    # -- regular state --

    def has_state(self, state: RuntimeRecordStateEnum) -> bool:
        return any([st for st in self.states if st.state_name == state])

    def get_state(self, state: RuntimeRecordStateEnum, first=True) -> RuntimeRecordState:
        states = [st for st in self.states if st.state_name == state]
        if first:
            return states[0]
        else:
            return states[-1]

    def add_state(self,
                  state: RuntimeRecordStateEnum,
                  time: float, tick: int,
                  state_values: TagValueCollection | None) -> RuntimeRecordState:
        record_state = RuntimeRecordState(state, time, tick, state_values)
        self.states.append(record_state)
        return record_state

    def add_state_awaiting_threshold(self, time: float, tick: int, state_values: TagValueCollection | None):
        self.add_state(RuntimeRecordStateEnum.AwaitingThreshold, time, tick, state_values)

    def add_state_awaiting_condition(self, time: float, tick: int, state_values: TagValueCollection | None):
        self.add_state(RuntimeRecordStateEnum.AwaitingCondition, time, tick, state_values)

    def add_state_awaiting_interrupt(self, time: float, tick: int, state_values: TagValueCollection | None):
        self.add_state(RuntimeRecordStateEnum.AwaitingInterrrupt, time, tick, state_values)

    def add_state_started(self, time: float, tick: int, start_values: TagValueCollection):
        self.add_state(RuntimeRecordStateEnum.Started, time, tick, start_values)

    def add_state_completed(self, time: float, tick: int, end_values: TagValueCollection):
        self.add_state(RuntimeRecordStateEnum.Completed, time, tick, end_values)

    def add_state_failed(self, time: float, tick: int, end_values: TagValueCollection):
        self.add_state(RuntimeRecordStateEnum.Failed, time, tick, end_values)

    def add_state_cancelled(self, time: float, tick: int, state_values: TagValueCollection | None):
        self.add_state(RuntimeRecordStateEnum.Cancelled, time, tick, state_values)

    def add_state_forced(self, time: float, tick: int, state_values: TagValueCollection | None):
        self.add_state(RuntimeRecordStateEnum.Forced, time, tick, state_values)

    def set_end_visit(self, time: float, tick: int, end_values: TagValueCollection):
        self.visit_end_time = time
        self.visit_end_tick = tick
        self.end_values = end_values

    # -- command state --

    def add_state_uod_command_set(self, uod_command: EngineCommand,
                                  time: float, tick: int,
                                  state_values: TagValueCollection | None) -> UUID:
        state = self.add_state(RuntimeRecordStateEnum.UodCommandSet, time, tick, state_values)
        state.command = uod_command
        state.command_exec_id = uuid4()
        return state.command_exec_id

    def add_state_internal_engine_command_set(
            self, command: EngineCommand,
            time: float, tick: int,
            state_values: TagValueCollection | None) -> UUID:
        state = self.add_state(RuntimeRecordStateEnum.InternalEngineCommandSet, time, tick, state_values)
        state.command = command
        state.command_exec_id = uuid4()
        return state.command_exec_id

    def add_command_state_cancelled(
            self, command_exec_id: UUID,
            time: float, tick: int, end_values: TagValueCollection):
        state = self.add_state(RuntimeRecordStateEnum.Cancelled, time, tick, end_values)
        state.command_exec_id = command_exec_id

    def add_command_state_started(
            self, command_exec_id: UUID,
            time: float, tick: int, end_values: TagValueCollection):
        state = self.add_state(RuntimeRecordStateEnum.Started, time, tick, end_values)
        state.command_exec_id = command_exec_id

    def add_command_state_completed(
            self, command_exec_id: UUID,
            time: float, tick: int, end_values: TagValueCollection):
        state = self.add_state(RuntimeRecordStateEnum.Completed, time, tick, end_values)
        state.command_exec_id = command_exec_id

    def add_command_state_failed(
            self, command_exec_id: UUID,
            time: float, tick: int, end_values: TagValueCollection):
        state = self.add_state(RuntimeRecordStateEnum.Failed, time, tick, end_values)
        state.command_exec_id = command_exec_id


class RuntimeRecordState():
    def __init__(
            self,
            state: RuntimeRecordStateEnum,
            time: float, tick: int,
            values: TagValueCollection | None
    ) -> None:
        self.state_name: RuntimeRecordStateEnum = state
        self.state_time: float = time
        self.state_tick: int = tick
        self.values: TagValueCollection | None = values
        self.command: EngineCommand | None = None
        self.command_exec_id: UUID | None = None

    def __str__(self) -> str:
        return f"{self.state_name}"


class RuntimeRecordStateEnum(StrEnum):
    """ Defines the states runtime records can take """

    Visited = auto()
    """ Instruction node was visited"""
    UodCommandSet = auto()
    """ Uod command was set """
    InternalEngineCommandSet = auto()
    """ Internal engine command was set """
    AwaitingThreshold = auto()
    """ Waiting for threshold """
    AwaitingCondition = auto()
    """ Waiting for condition """
    AwaitingInterrrupt = auto()
    """ Instruction is awaiting invocation by interrupt """
    Started = auto()
    """ Command has started """
    Cancelled = auto()
    """ Command was cancelled. Either by overlapping command or explicitly by user """
    Forced = auto()
    """ Waiting command was forcibly started by user """
    Completed = auto()
    """ Command has completed """
    Failed = auto()
    """ Command has failed """

    @staticmethod
    def has_value(value: str):
        """ Determine if enum has this string value defined. Case sensitive. """
        return hasattr(RuntimeRecordStateEnum, value)


class RunLog():
    def __init__(self) -> None:
        self.id: str = ""
        self.items: list[RunLogItem] = []

    def size(self) -> int:
        return len(self.items)


class RunLogItem():
    def __init__(self) -> None:
        self.id: str = ""
        """ Exec_id of the command """
        self.name: str = ""
        self.start: float = 0
        self.end: float | None = None
        self.state: RunLogItemState = RunLogItemState.Waiting
        self.progress: float | None = None
        self.start_values: TagValueCollection = TagValueCollection.empty()
        self.end_values: TagValueCollection = TagValueCollection.empty()
        self.forcible: bool = False
        self.forced: bool = False
        self.cancellable: bool = False
        self.cancelled: bool = False


class RunLogItemState(StrEnum):
    """ Defines the states that commands in the run log can take """
    Waiting = auto()
    """ Waiting for threshold or condition """
    Started = auto()
    """ Command has started """
    Cancelled = auto()
    """ Command was cancelled. Either by overlapping command or explicitly by user """
    Forced = auto()
    """ Waiting command was forcibly started by user """
    Completed = auto()
    """ Command has completed """
    Failed = auto()
    """ Command failed """

    @staticmethod
    def has_value(value: str):
        """ Determine if enum has this string value defined. Case sensitive. """
        return hasattr(RunLogItemState, value)



"""add engine_version to recent_run

Revision ID: ec559277c8da
Revises: 72449425de7c
Create Date: 2024-02-02 11:55:49.427312

"""
from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op
from openpectus import __version__

# revision identifiers, used by Alembic.
revision: str = 'ec559277c8da'
down_revision: Union[str, None] = '72449425de7c'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentRuns', schema=None) as batch_op:
        batch_op.add_column(sa.Column('engine_version', sa.String(), nullable=False, server_default=__version__))
        batch_op.alter_column('engine_version', server_default=None)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentRuns', schema=None) as batch_op:
        batch_op.drop_column('engine_version')

    # ### end Alembic commands ###



from __future__ import annotations

from datetime import datetime
from enum import StrEnum, auto
from typing import Dict

from openpectus.aggregator.models import Method, MethodState, PlotConfiguration, RunLog, AggregatedErrorLog
from sqlalchemy import JSON, ForeignKey, MetaData
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship, attribute_keyed_dict


class DBModel(DeclarativeBase):
    """ Base model for data entity classes. """
    id: Mapped[int] = mapped_column(primary_key=True)
    metadata = MetaData(naming_convention={
        "ix": "ix_%(column_0_label)s",
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ck": "ck_%(table_name)s_`%(constraint_name)s`",
        "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
        "pk": "pk_%(table_name)s"
    })


class RecentEngine(DBModel):
    __tablename__ = "RecentEngines"

    engine_id: Mapped[str] = mapped_column(unique=True)
    run_id: Mapped[str | None] = mapped_column()
    run_started: Mapped[datetime | None] = mapped_column()
    run_stopped: Mapped[datetime | None] = mapped_column()
    name: Mapped[str] = mapped_column()
    system_state: Mapped[str] = mapped_column()
    location: Mapped[str] = mapped_column()
    last_update: Mapped[datetime] = mapped_column()
    contributors: Mapped[list[str]] = mapped_column(type_=JSON, default=[])
    required_roles: Mapped[set[str]] = mapped_column(type_=JSON, default=[])


class RecentRun(DBModel):
    """ Represents a historical run of a process unit. """
    __tablename__ = "RecentRuns"
    engine_id: Mapped[str] = mapped_column()
    run_id: Mapped[str] = mapped_column()
    engine_computer_name: Mapped[str] = mapped_column()
    engine_version: Mapped[str] = mapped_column()
    engine_hardware_str: Mapped[str] = mapped_column()
    aggregator_computer_name: Mapped[str] = mapped_column()
    aggregator_version: Mapped[str] = mapped_column()
    uod_name: Mapped[str] = mapped_column()
    uod_filename: Mapped[str] = mapped_column()
    uod_author_name: Mapped[str] = mapped_column()
    uod_author_email: Mapped[str] = mapped_column()
    started_date: Mapped[datetime] = mapped_column()
    completed_date: Mapped[datetime] = mapped_column()
    contributors: Mapped[list[str]] = mapped_column(type_=JSON, default=[])  # really a set but this cannot be mapped?
    required_roles: Mapped[list[str]] = mapped_column(type_=JSON, default=[])  # really a set but this cannot be mapped?


class RecentRunMethodAndState(DBModel):
    __tablename__ = "RecentRunMethodAndStates"
    run_id: Mapped[str] = mapped_column()
    method: Mapped[Method] = mapped_column(type_=JSON)
    state: Mapped[MethodState] = mapped_column(type_=JSON)


class RecentRunRunLog(DBModel):
    __tablename__ = "RecentRunRunLogs"
    run_id: Mapped[str] = mapped_column()
    run_log: Mapped[RunLog] = mapped_column(type_=JSON)


class RecentRunErrorLog(DBModel):
    __tablename__ = "RecentRunErrorLogs"
    run_id: Mapped[str] = mapped_column()
    error_log: Mapped[AggregatedErrorLog] = mapped_column(type_=JSON)


class RecentRunPlotConfiguration(DBModel):
    __tablename__ = "RecentRunPlotConfigurations"
    run_id: Mapped[str] = mapped_column()
    plot_configuration: Mapped[PlotConfiguration] = mapped_column(type_=JSON)


class PlotLogEntryValue(DBModel):
    __tablename__ = "PlotLogEntryValues"
    plot_log_entry_id: Mapped[int] = mapped_column(ForeignKey('PlotLogEntries.id'))
    tick_time: Mapped[float] = mapped_column()
    value_str: Mapped[str | None] = mapped_column()
    value_float: Mapped[float | None] = mapped_column()
    value_int: Mapped[int | None] = mapped_column()

    @property
    def value(self):
        if self.value_int is not None:
            return self.value_int
        if self.value_float is not None:
            return self.value_float
        if self.value_str is not None:
            return self.value_str


ProcessValueValueType = str | float | int | None


class ProcessValueType(StrEnum):
    STRING = auto()
    FLOAT = auto()
    INT = auto()
    CHOICE = auto()
    NONE = auto()


def get_ProcessValueType_from_value(value: ProcessValueValueType) -> ProcessValueType:
    if value is None:
        return ProcessValueType.NONE
    if isinstance(value, str):
        return ProcessValueType.STRING
    elif isinstance(value, int):
        return ProcessValueType.INT
    elif isinstance(value, float):
        return ProcessValueType.FLOAT
    else:
        raise ValueError("Invalid value type: " + type(value).__name__)


class PlotLogEntry(DBModel):
    __tablename__ = "PlotLogEntries"
    name: Mapped[str] = mapped_column()
    values: Mapped[list[PlotLogEntryValue]] = relationship(cascade="all, delete-orphan")
    value_unit: Mapped[str | None] = mapped_column()
    value_type: Mapped[ProcessValueType] = mapped_column()
    plot_log_id: Mapped[int] = mapped_column(ForeignKey('PlotLogs.id'))
    plot_log: Mapped[PlotLog] = relationship()


class PlotLog(DBModel):
    __tablename__ = "PlotLogs"
    engine_id: Mapped[str] = mapped_column()
    run_id: Mapped[str] = mapped_column()
    entries: Mapped[Dict[str, PlotLogEntry]] = relationship(
        collection_class=attribute_keyed_dict("name"),
        back_populates='plot_log',
        cascade="all, delete-orphan"
    )



import os
import subprocess
import unittest


class TestValidateDemoUOD(unittest.TestCase):
    def test_validate(self):
        # Construct path to file relative to test_validate_demo_uod.py
        uod_file_path = os.path.join(
            os.path.dirname(  # openpectus
                os.path.dirname(  # openpectus/test
                    os.path.dirname(__file__)  # openpectus/test/engine
                )
            ),
            'engine',  # openpectus/engine
            'configuration',  # openpectus/engine/configuration
            'demo_uod.py',  # openpectus/engine/configuration/demo_uod.py
        )

        # Run validation in subprocess to avoid side effects from other
        # unittests messing with the logging module.
        process = subprocess.run(
            ["pectus-engine", "-v", "-uod", uod_file_path],
            capture_output=True,
            encoding="utf-8"
        )

        # Look for errors or worse in the logs
        if "ERROR:openpectus" in process.stderr:
            print(process.stderr)
            self.assertTrue(
                False,
                msg="Error encountered during validation of 'demo_uod.py'."
            )


if __name__ == "__main__":
    unittest.main()



from types import NoneType
from typing import Any

import openpectus.aggregator.routers.dto as namespace
from pydantic import BaseModel

namespace_name = namespace.__name__
namespace_types = dir(namespace)


def serialize(msg: BaseModel) -> dict[str, Any]:
    """ Serialize a dto message in a round-trippable fashion. """
    json_dict = msg.model_dump()
    t = type(msg)
    qualified_type_name = t.__qualname__
    assert t.__module__ == namespace_name, \
        f"Cannot serialize type '{qualified_type_name}' which is not defined in namespace {namespace_name}"
    json_dict["_type"] = qualified_type_name
    return json_dict


def deserialize(json_dict: dict[str, Any]) -> BaseModel:
    """ Deserializes a dto message to the proper concrete BaseModel subtype.

    Supports messages serialized to dict by the serialize() function.
    """
    if "_type" not in json_dict.keys():
        raise ValueError("Deserialization error. Key '_type' missing.")
    message_type = json_dict["_type"]

    cls = try_get_ctor_from_namespace(message_type, namespace=namespace)

    assert not isinstance(cls, NoneType), \
        f"Class name '{cls}' cannot be deserialized since it is not defined in namespace '{namespace_name}'."
    try:
        msg = cls(**json_dict)
        assert isinstance(msg, BaseModel)
        return msg
    except Exception:
        raise


def try_get_ctor_from_namespace(type_name: str, namespace):
    if "." in type_name:  # Nested type
        outer_name, inner_name = type_name.split(".")[0:2]
        if "." in inner_name:
            raise NotImplementedError("Nested inner type not supported")
        if outer_name not in namespace_types:
            ValueError(f"Nested type '{type_name}' cannot be deserialized because the outer type {outer_name} is not " +
                       f"defined in namespace {namespace_name}")
        outer_type = getattr(namespace, outer_name, None)
        if outer_type is None:
            raise ValueError(f"Failed to get outer class {outer_name} of nested type {type_name}")
        return getattr(outer_type, inner_name, None)
    else:
        return getattr(namespace, type_name, None)



from pydantic import BaseModel

# Topics

# No topics in aggregator-engine protocol, rest and rpc only


# Messages

class MessageBase(BaseModel):
    """ Base class for protocol messages for both REST and WebSocket communication.

    These inherit from pydantic BaseModel to support automatic (de-)serialization when sent over
    fastapi_websocket_pubsub. Additionally they support automatic openapi schema generation for
    use with REST.

    Unfortunately, for websocket rpc, (de-)serialization is not automatic but is handled using
    the openpectus.protocol.serialization module.
    """
    version: int = 0
    """ Protocol is currently unversioned. This field is included to support future versioning. """


class SuccessMessage(MessageBase):
    """ Indicates operation success"""
    pass


class ErrorMessage(MessageBase):
    """ Returned whenever an error occurs"""
    message: str | None = None
    exception_message: str | None = None


class ProtocolErrorMessage(ErrorMessage):
    protocol_msg: str


RpcErrorMessage = ErrorMessage | ProtocolErrorMessage
RpcStatusMessage = SuccessMessage | RpcErrorMessage


# RpcValueMessage = MessageBase | RpcErrorMessage
